// Copyright 2011 NLX - Natural Language and Speech Group
// of the University of Lisbon
//
// This file is part of lkb2standard.
//
// lkb2standard is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lkb2standard is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lkb2standard. If not, see <http://www.gnu.org/licenses/>.

package pt.ul.fc.di.nlx.lkb2standard;

import java.util.Properties;

import jargs.gnu.CmdLineParser;

/**
 * The Simplifier is a command line tool that transforms/simplifies the phrase structure trees that have been exported by the grammar.
 * <p>
 * The class itself is only concerned with the command line interface.
 * It handles the parsing of command line parameters and the setting of program properties.
 * <p>
 * The file handling is performed by the <code>Traverser</code> class.<br>
 * The tree manipulations are performed by the <code>TreeHandler</code> class. 
 * <p>
 * TODO: Rename class to lkb2standard.
 *  
 * @author jsilva
 * 
 */
public class Simplifier {
	
	public static void main(String args[]) {
		
		// Defines the syntax of the command line parameters
		CmdLineParser cmdLineParser = new CmdLineParser();
		CmdLineParser.Option optConfigFile = cmdLineParser.addStringOption('c', "config"); 
		CmdLineParser.Option optVersion = cmdLineParser.addIntegerOption('v', "version"); 
		
		// Parses the command line
		try {
			cmdLineParser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(2);
		}
		String configFile = (String) cmdLineParser.getOptionValue(optConfigFile, "conf/simplifier.properties");
		Integer corporaVersion = (Integer) cmdLineParser.getOptionValue(optVersion); // TODO: Mandatory argument? Default to latest version?
		
		// Sets default and overriding properties
		Properties properties = new Properties();
		initializeProperties(properties, configFile);
		
		// Builds the base path for the current version of the data
		String versionBaseDirName =	properties.getProperty("REPOSITORY") + "/" + properties.getProperty("CORPORA") + "/" + "version" + corporaVersion + "/";
		
		// Adjudication directory (exported parses)
		String adjudDirname = properties.getProperty("ADJUDICATION");
		if (!adjudDirname.startsWith("/")) {
			adjudDirname = versionBaseDirName + adjudDirname;
		}
		System.out.println("Path to adjudication directory:\n" + adjudDirname + "\n");
		properties.setProperty("_ADJUDDIRNAME", adjudDirname);
		
		// CINTIL directory (annotated sentences)
		String cintilDirname = properties.getProperty("CINTIL");
		if (!cintilDirname.startsWith("/")) {
			cintilDirname = versionBaseDirName + cintilDirname;
		}
		System.out.println("Path to CINTIL directory:\n" + cintilDirname + "\n");
		properties.setProperty("_CINTILDIRNAME", cintilDirname);
		
		// Grammar lexicon
		String lexiconFilename = properties.getProperty("LEXICON");
		if (!lexiconFilename.startsWith("/")) {
			lexiconFilename = versionBaseDirName + lexiconFilename;
		}
		System.out.println("Path to grammar lexicon file:\n" + lexiconFilename + "\n");
		properties.setProperty("_LEXICONFILENAME", lexiconFilename);
		
		// Output directory
		String outputDirname = properties.getProperty("OUTPUT");
		if (outputDirname.isEmpty()) {
			System.out.println("Suppress writing to output directory.");
		} else {
			if (!outputDirname.startsWith("/")) {
				outputDirname = versionBaseDirName + outputDirname;
			}
			System.out.println("Path to output directory:\n" + outputDirname + "\n");
		}
		properties.setProperty("_OUTPUTDIRNAME", outputDirname);
		
		// Traverse test suite directories
		// All relevant information is passed through Properties at object creation.
		Traverser traverser = new Traverser(properties);
		try {
			traverser.traverseDirectories();
		} catch (Exception e) {
			System.err.println(e.getMessage());
			System.exit(1);
		}
	}	
	
	/**
	 * Prints the command line usage instructions.
	 */
	private static void printUsage() {
		System.err.println(
				"Usage: Simplifier [OPTION]\n" +
				"\t-v, --version\n\t\ttreebank version\n" +
				"\t-c, --config\n\t\tconfig file");
	}
		
	/**
	 * Initializes the Properties object. Starts by setting default values and
	 * then reads the .properties file for any new or overriding settings.
	 * 
	 * @param properties a Java Properties object where properties will be stored
	 * @param path       path to the .properties file
	 * @see              java.util.Properties
	 */
	private static void initializeProperties(Properties properties, String path) {

		// Sets default values
		properties.setProperty("CORPORA",
				"semanticShare/annotationWorkbench/annotatedCorpora");
		properties.setProperty("ADJUDICATION",
				"annotation/treebank/adjudication/exportAdjud");
		properties.setProperty("CINTIL",
				"evolvingCorpora");
		properties.setProperty("LEXICON",
				"grammar/src/lexicon.tdl");
		properties.setProperty("OUTPUT",
				"");
		properties.setProperty("SSTREEB",
				"^SSTreeB-\\S+-\\d+-V\\d+-EXPORT$");
		properties.setProperty("SUITEGZ",
				"^\\d+.gz$");
		properties.setProperty("SKIPSSTREEB",
				"");
		properties.setProperty("SKIPSUITEGZ",
				"");
		properties.setProperty("CINTILFILE",
				"items.txt");
		properties.setProperty("PUNCTUATIONNODE",
				"PNT");

		// Reads .properties file, overriding default values if necessary
		try {
			properties.load(new java.io.FileInputStream(path));
		} catch (java.io.FileNotFoundException e) {
			System.err.println("Configuration file not found.");
		} catch (java.io.IOException e) {
			System.err.println("Error accessing configuration file: "
					+ e.getMessage());
		}
		
	}
	
}
