// Copyright 2011 NLX - Natural Language and Speech Group
// of the University of Lisbon
//
// This file is part of lkb2standard.
//
// lkb2standard is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lkb2standard is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lkb2standard. If not, see <http://www.gnu.org/licenses/>.

package pt.ul.fc.di.nlx.lkb2standard;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Properties;

import pt.ul.fc.di.nlx.common.conversor.io.file.SentenceFile;
import pt.ul.fc.di.nlx.common.conversor.io.file.SentenceSuiteInfoFile;

/**
 * This class bridges the gap between the filesystem and tree data structures.
 * It is responsible for finding and scanning the treebank export directories and extracting the trees from the files therein.
 * <p>
 * There are three important locations:
 * <ul>
 * <li>The tree directory contains the Redwoods files exported by the grammar. Each test suite directory holds one gzip file for each sentence in that test suite. The name of the file is the number of the sentence.
 * <li>The annotated sentence directory contains an items.txt file for each test suite. This file in turn contains all sentences in that test suite, in plain text and annotated in the CINTIL format.
 * <li>The output directory is where the resulting trees will be placed.
 * </ul>
 * <p>
 * The tree manipulations are performed by the <code>TreeHandler</code> class.<br>
 * The code that handles the extraction of the trees from the files is in the <code>common.conversor.io.file</code> package.
 * <p>
 * TODO: Rename the variables and Properties fields, for clarity. For instance, "_ADJUDDIRNAME" should be "_EXPORTDIRNAME", or something similar, since it runs over any Redwoods export directory and not necessarily the ajudicator's export.
 * TODO: A better progress report (percentage, progress bar, etc.)
 * 
 * @author jsilva
 *
 */
public class Traverser {
	
	Properties properties;
	TreeHandler treeHandler;
	
	/**
	 * Create a new Traverser object.
	 * <p>
	 * All relevant information (directories, files, etc.) is passed in a java.util.Properties object.
	 * 
	 * @param properties
	 */
	public Traverser(Properties properties) {
		this.properties = properties;
		treeHandler = new TreeHandler(properties);
	}

	/**
	 * Goes through each test suite directory and the files within.
	 * 
	 * @param properties
	 * @throws FileNotFoundException 
	 */
	public void traverseDirectories() throws FileNotFoundException {

		File adjudDir = new File(properties.getProperty("_ADJUDDIRNAME"));
		if (!adjudDir.exists() || !adjudDir.isDirectory()) {
			throw new FileNotFoundException("Adjudication directory not found: " + adjudDir);
		}
		
		File cintilDir = new File(properties.getProperty("_CINTILDIRNAME"));
		if (!cintilDir.exists() || !cintilDir.isDirectory()) {
			throw new FileNotFoundException("CINTIL directory not found: " + cintilDir);
		}
		
		String outputDirname = properties.getProperty("_OUTPUTDIRNAME");
		boolean writeOutputFiles = !outputDirname.isEmpty();
		File outputDir = null;
		if (writeOutputFiles) {
			outputDir = new File(outputDirname);
			if (!outputDir.exists() || !outputDir.isDirectory()) {
				throw new FileNotFoundException("Output directory not found: " + outputDir);
			}
		}
		
		// Filters for selection
		RegexpFilter ssTreeBFilter = new RegexpFilter(properties.getProperty("SSTREEB"));
		RegexpFilter suiteGZFilter = new RegexpFilter(properties.getProperty("SUITEGZ"));
		
		// Filters for exclusion
		RegexpFilter skip_ssTreeBFilter = properties.getProperty("SKIPSSTREEB").equalsIgnoreCase("") ? null : new RegexpFilter(properties.getProperty("SKIPSSTREEB"));
		RegexpFilter skip_suiteGZFilter = properties.getProperty("SKIPSUITEGZ").equalsIgnoreCase("") ? null : new RegexpFilter(properties.getProperty("SKIPSUITEGZ"));
		
		if (adjudDir.isDirectory()) {
			File[] filesInAdjudDir = adjudDir.listFiles(ssTreeBFilter);
			java.util.Arrays.sort(filesInAdjudDir); // sorts in-place
			double count = 0;
			for (File subdirectory : filesInAdjudDir) {
				boolean skipDir = (skip_ssTreeBFilter != null) ? skip_ssTreeBFilter.accept(subdirectory.getParentFile(), subdirectory.getName()) : false;
				if (subdirectory.isDirectory() && !skipDir) {
					String cintilSubdir = sstreebDir2cintilDir(subdirectory.getName());
					String outputSubdir = sstreebDir2outputDir(subdirectory.getName());

					// FIXME: Fix for directory with wrong names in v3
					// TODO: Make it so that this fix is only applied to v3
					if (cintilSubdir.contains("cintilSS-")) {
						cintilSubdir = cintilSubdir.replace("-V03", "-V04");
					}
					
					File cintilFile = new File(new File(cintilDir, cintilSubdir), properties.getProperty("CINTILFILE"));
					SentenceSuiteInfoFile ssinfof = new SentenceSuiteInfoFile(cintilFile.getParent());
					
					if (ssinfof.readFile()) {
						String infoHeader = subdirectory.getName() + " (" + cintilSubdir + "/" + cintilFile.getName() + ")";
						char[] bar = new char[infoHeader.length()];
						java.util.Arrays.fill(bar, '-');
						System.out.println("\n" + String.valueOf(bar) + "\n" + infoHeader);
						File[] gzFiles = subdirectory.listFiles(suiteGZFilter);
						java.util.Arrays.sort(gzFiles); // sorts in-place
						for (File adjudFile : gzFiles) {
							SentenceFile sf = new SentenceFile(adjudFile.getAbsolutePath());
							boolean skipGZ = (skip_suiteGZFilter != null) ? skip_suiteGZFilter.accept(adjudFile.getParentFile(), adjudFile.getName()) : false;
							if (sf.readFile() && !skipGZ) {
								File outputFile = null;
								if (writeOutputFiles) {
									outputFile = new File(new File(outputDir, outputSubdir), sstreebFile2outputFile(adjudFile.getName()));
									System.out.println("\n" + adjudFile.getName() + " => " + outputSubdir + "/" + outputFile.getName());
								} else {
									System.out.println("\n" + adjudFile.getName());
								}
//								System.out.println("\nHeader:\n" + subdirectory.getName() + "/" + adjudFile.getName());
								long startms = System.currentTimeMillis();
								String result =	processFiles(sf, ssinfof);
								long endms = System.currentTimeMillis();
								System.out.println("\nResult:\n" + result);
								if (writeOutputFiles) {
									writeToOutput(outputFile, result);
								}
								System.out.println("\nElapsed: " + (endms - startms) + " ms");
							}
						}
					}
				}
				count++;
				//System.out.printf("%3.0f%% completed\n", (count/filesInAdjudDir.length)*100);
			}
		}
	}

	/**
	 * 
	 * @param outputFile
	 * @param result
	 */
	public void writeToOutput(File outputFile, String result) {
		File parentPath = outputFile.getParentFile();

		try {
			if (parentPath.mkdir()) {
				if (outputFile.createNewFile()) {
					// Created file
				} else {
					// File already exists?
				}
			}
		} catch (Exception e) {
			// TODO: handle exception
		}

		try {
			java.io.FileWriter fw = new java.io.FileWriter(outputFile);
			fw.write(result + "\n");
			fw.close();			
		} catch (Exception e) {
			// TODO: handle exception
		}

	}
	
	/**
	 * Extracts the strings that represent the phrase structure tree, the derivation tree and the annotated sentence.
	 * It then runs the transformation pipeline.
	 * 
	 * @param sf      the wrapper for the trees exported by the grammar 
	 * @param ssinfof the wrapper for the annotated sentences file
	 */
	public String processFiles(SentenceFile sf, SentenceSuiteInfoFile ssinfof) {
		
		String result = "";
		
		// Gets derivation tree
		String ruleTreeStr = sf.getCompleteTree();

		// Gets phrase structure tree
		String treeStr = sf.getIncompleteTree();

		// Reads sentence from CINTIL file
		String annotated = null;
		java.util.regex.Pattern p = java.util.regex.Pattern.compile("\\[([0-9]+)\\]");
		java.util.regex.Matcher m = p.matcher(sf.getHeader());
		if (m.find()) {
			String id = m.group(1);
		
			// FIXME: Temporary hack to account for a mismatch in the numbering of the test suites in v2
			// TODO: Make it so that this fix is only applied to v2
//			System.out.println("Header (sf): " + sf.getHeader());		
//			if (sf.getHeader().contains("SSTreeB-aTSTS")) {
//				System.out.println("ID (original): " + id);
//				Integer num = new Integer(id);
//				num = num + 17764;
//				id = num.toString();
//				System.out.println("ID (adjusted): " + id);
//			}
			
			annotated = ssinfof.getAnnotated(id);
			
			System.out.println("\nSentence:\n" + ssinfof.getSentence(id));
//			System.out.println(annotated);
//			System.out.println(treeStr);
			
			result = treeHandler.runPipeline(treeStr, ruleTreeStr, annotated);
		}
		
		return result;

	}
	

	/**
	 * Returns the name of the CINTIL directory that corresponds to the given test suite directory.
	 * 
	 * @param  dirname the test suite directory name
	 * @return         the corresponding CINTIL directory name
	 */
	private String sstreebDir2cintilDir(String dirname) {
		// TODO: Implement a more flexible, version-dependent, external, user-defined system. Fail when no case is applicable?.
		if (dirname.contains("aTSTS")) {
			dirname = dirname.replaceFirst("SSTreeB-aTSTS", "/testSuites/partTestSuites/SSTreeB-aTSTS").replaceFirst("-EXPORT", "");
		} else if (dirname.contains("bCINT")) {
			dirname = dirname.replaceFirst("SSTreeB-bCINT", "/cintil/subCintil/cintilSS").replaceFirst("-EXPORT", "");
		} else if (dirname.contains("cCINT")) {
			dirname = dirname.replaceFirst("SSTreeB-cCINT", "/cintil/subCintil/cintilSSc").replaceFirst("-EXPORT", "");
		} else if (dirname.contains("dPENN")) {
			dirname = dirname.replaceFirst("SSTreeB-dPENN", "/PennTreebankPOR/subPennTreebank/PennTreebank").replaceFirst("-EXPORT", "");
		} else {
			dirname = dirname.replaceFirst("SSTreeB-eCTMP", "/CETEMPublico/subCETEMPublico/SSTreeB-eCTMP").replaceFirst("-EXPORT", "");
		}
		return dirname;
	}
	
	
	/**
	 * Returns the name of the output directory that corresponds to the given test suite directory.
	 * 
	 * @param  dirname the test suite directory name
	 * @return         the corresponding output directory name
	 */
	private String sstreebDir2outputDir(String dirname) {
		// TODO: Implement a more flexible, version-dependent, external, user-defined system. Fail when no case is applicable?.
		return dirname;
	}

	/**
	 * Returns the name of the output file that corresponds to the given test suite file.
	 * 
	 * @param filename the test suite filename
	 * @return
	 */
	private String sstreebFile2outputFile(String filename) {
		// TODO: Implement a more flexible, version-dependent, external, user-defined system. Fail when no case is applicable?.
		return filename.replaceFirst("\\.gz", "");
	}
}
