// Copyright 2011 NLX - Natural Language and Speech Group
// of the University of Lisbon
//
// This file is part of lkb2standard.
//
// lkb2standard is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lkb2standard is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lkb2standard. If not, see <http://www.gnu.org/licenses/>.

package pt.ul.fc.di.nlx.lkb2standard;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Properties;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.stanford.nlp.ling.Label;
import edu.stanford.nlp.ling.StringLabelFactory;
import edu.stanford.nlp.trees.LabeledScoredTreeFactory;
import edu.stanford.nlp.trees.PennTreeReader;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TreeReader;
import edu.stanford.nlp.trees.tregex.ParseException;
import edu.stanford.nlp.trees.tregex.TregexMatcher;
import edu.stanford.nlp.trees.tregex.TregexPattern;
import edu.stanford.nlp.trees.tregex.tsurgeon.Tsurgeon;
import edu.stanford.nlp.trees.tregex.tsurgeon.TsurgeonPattern;


/**
 * This class holds all methods that transform the tree.
 * <p>
 * TODO: Generate Javadoc (putting it in the correct directory) and commit it to repository.
 *   
 * @author jsilva
 *
 */
public class TreeHandler {
	
	Properties properties;
	
	// Objects that, though only used in their own methods, are reused for every parse and thus need only be created once
	HashMap<String, String> entry2type;
	LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
	
	// Matches quoted strings, while allowing for escaped quotes (vd. Mastering regular expressions, pg. 198)
	Pattern quotedString = Pattern.compile("\\\"((\\\\.|[^\\\\\"])*)\\\"");

	
	/**
	 * Create a new TreeHandler object.
	 * 
	 * @param properties
	 */
	public TreeHandler(Properties properties) {
		this.properties = properties;
		try {
			entry2type = readLexicon();
		} catch (IOException e) {
			System.err.println(e.getMessage());
			System.exit(1);
		}
	}
	
	
	/**
	 * Reads a Redwoods lexicon into a hash that maps lexical entries into lexical types.
	 * 
	 * @return A hash that maps lexical entries into lexical types.
	 * @throws FileNotFoundException
	 */
	private HashMap<String, String> readLexicon() throws IOException {
		BufferedReader reader = new BufferedReader(new FileReader(properties.getProperty("_LEXICONFILENAME")));
		HashMap<String, String> hash = new HashMap<String, String>();
		while (reader.ready()) {
			String line = reader.readLine().trim();
			if (line.endsWith(":=")) {
				String entry = line.replaceFirst(" *:=", "").toLowerCase();
				String type = reader.readLine().trim().replaceFirst(" *&.*", "").toLowerCase().replaceAll("[*+]", "~");
				hash.put(entry, type);
//				System.out.println(entry + " -> " + type);
			}
		}
		return hash;
	}
	

	/**
	 * A convenient method to allow selecting only those trees matching a pattern.  
	 * 
	 * @param tree
	 * @return
	 */
	private boolean filterTree(Tree tree) {
		boolean result = false;
		try {
			TregexPattern tregex = TregexPattern.compile("/MISSING-(?:COMPLEMENTIZER|TEMPORAL-PREPOSITION)/");
			result = tregex.matcher(tree).find();
		} catch (ParseException e) {
			e.printStackTrace();
		}
		return result;
	}
	
	/**
	 * Runs the transformation pipeline.
	 * <p>
	 * Methods that decorate the tree with information on empty categories (which will be required by other methods):
	 * <ul>
	 * <code>annotateNullSJNodes</code>
	 * <code>annotateEllipsisNodes</code>
	 * <code>annotateModifierNodes</code>
	 * <code>annotateGapNodes</code>
	 * </ul>
	 * <p>
	 * Methods that decorate the tree with information on raising/control/etc. (which will be required by other methods):
	 * <code>annotateSubjectRaising</code>
	 * <code>annotateCopula</code>
	 * <code>annotateAnticausative</code>
	 * <code>annotateControl</code>
	 * <p> 
	 * The standard pipeline of transformations:
	 * <ul>
	 * <li><code>expandMultiwordLeaves</code>
	 * <li><code>detachPunctuation</code>
	 * <li><code>restoreOrthography</code>
	 * <li><code>flattenUnaryChains</code>
	 * <li><code>exciseMultiwordNodes</code>
	 * <li><code>moveCoordinationComma</code>
	 * <li><code>moveNoConjCoordComma</code>
	 * <li><code>promotePunctuation</code>
	 * <li><code>addPhrasalNodes</code>
	 * <li><code>addAnnotationFromSentence</code>
	 * </ol>
	 * Methods that add empty categories to the tree:
	 * <ul>
	 * <li><code>addNullSJBranch</code>
	 * <li><code>addNullVerbBranch</code>
	 * <li><code>addEllipsisBranch</code>
	 * <li><code>addGapBranch</code>
	 * </ul>
	 * <p>
	 * 
	 * @param psTreeStr     the phrase structure, as a string (from Redwoods export)
	 * @param derivTreeStr  the derivation tree, as a string (from Redwoods export)
	 * @param annotated     the annotated sentence (CINTIL/LX-Suite format)
	 */
	public String runPipeline(String psTreeStr, String derivTreeStr, String annotated) {
		
//		System.out.println("\nConstituent tree (raw string):\n" + psTreeStr);
//		System.out.println("\nDerivation tree (raw string):\n" + derivTreeStr);
//		System.out.println("\nAnnotated sentence (raw string):\n" + annotated);
		
		// Preprocess input strings and create the constituent tree and the derivation tree
		Tree constituentTree = string2tree(preprocessTreeString(renameParenthesisInLeaves(psTreeStr, derivTreeStr)));
		Tree derivationTree = string2tree(preprocessTreeString(renameParenthesisInLeaves(derivTreeStr, derivTreeStr)));
		String annotatedSentence = fixBadTokenizationInSentence(annotated);
		
		// Sanity check: Ensure that the trees have matching structures
		if (!sameStructure(constituentTree, derivationTree)) {
			System.err.println("\nERROR: Constituency and derivation trees have different structures." +
					"\nConstituent tree (raw string):\n" + psTreeStr +
					"\nDerivation tree (raw string):\n" + derivTreeStr +
					"\nConstituent tree:\n" + constituentTree +
					"\nDerivation tree:\n" + derivationTree);
			System.exit(1);
		}
		
		// Only process trees matching a pattern
//		if (!filterTree(derivationTree)) {
//			return null;
//		}
		
//		System.out.println("\nConstituent tree:\n" + constituentTree);
//		System.out.println("\nDerivation tree:\n" + derivationTree);
		System.out.println("\nAnnotated sentence:\n" + annotated);
		
		// Uncomment to skip the remaining code
//		if (true) {
//			return null;
//		}

		// Fixes for v3
		fix_PercentNodes(constituentTree);
		fix_EntreENodes(constituentTree);
		
		// Annotates the constituent tree with information on empty categories and modifiers
		annotateNullSJNodes(constituentTree, derivationTree);
		annotateEllipsisNodes(constituentTree, derivationTree);
		annotateModifierNodes(constituentTree);
		annotateGapNodes(constituentTree, derivationTree);
		
		// Annotates the constituent tree with information regarding raising/copula/control
//		System.out.println("\nIntermediate tree:\n" + constituentTree);
		annotateSubjectRaising(constituentTree, derivationTree);
		annotateCopula(constituentTree, derivationTree);
		annotateAuxiliar(constituentTree, derivationTree);
		annotateAnticausative(constituentTree, derivationTree);
		annotateControl(constituentTree, derivationTree);

		// Annotate pre-terminals with their lexical type so that, later on, it can be added to the leaf as part of a feature bundle
		annotateWithLexicalTypes(constituentTree, derivationTree);

		// Fixing leaves
		replaceLeavesInConstituentTree(constituentTree, derivationTree);
		expandContractionsInClosedClassMWUNodes(constituentTree, derivationTree);
		
		// Fix for v3 - Here because it changes the tree, and would break derivation-constituency alignment if done before
		fix_Participles(constituentTree, derivationTree);

		// Aligning/matching leaves to tokens
		expandMultiwordLeaves(constituentTree);
		detachPunctuation(constituentTree, annotatedSentence);
		
		// Sanity check: Ensure that the tree and the annotated sentence have the same number of leafs/tokens
		if (!sameLength(constituentTree, annotatedSentence)) {
			System.err.println("\nERROR: Mismatch in the number of leafs and tokens." +
					"\nConstituent tree:\n" + constituentTree +
					"\nAnnotated sentence:\n" + annotatedSentence);
			System.exit(1);
		}
		
		// If we want to use/incorporate information from the annotated sentence, we should do it while leafs and tokens are aligned
		restoreOrthography(constituentTree, annotatedSentence);
		boolean addFeatureBundles = true;
		if (addFeatureBundles) {
			addFeatureBundles(constituentTree, annotatedSentence);
			addLexicalTypes(constituentTree);
		}
		
		// Excising unary chains
		// NB: Excising unary chains alters the structure that allows detecting coordinations that have no comma. Accordingly, these structures must be pre-processed before excising unary chains.
		relabelUnaryCP(constituentTree);
		moveNoConjCoordinationComma(constituentTree);
		flattenUnaryChains(constituentTree);
		
		// Moving and raising punctuation
		preprocessParentheticalModifierPunctuation(constituentTree);
		moveCoordinationComma(constituentTree);
		raisePunctuation(constituentTree);
		removeAnnotation(constituentTree, "M");
		
		// Handling MWUs
//		System.out.println("\nIntermediate tree:\n" + constituentTree);
		boolean mergeMWUnodes = true;
		boolean replaceIOBByTypes = false;
		boolean flattenCARDs = true;
		fixPRS2(constituentTree);
		if (mergeMWUnodes) {
			mergeMultiwordNodes(constituentTree, "^(?:[^*-/]+)[*-/].*$");
			if (replaceIOBByTypes) {
				replaceIOBByTypes(constituentTree, annotatedSentence);
			}
			if (flattenCARDs) {
				flattenCARDs(constituentTree);
			}
			fixSpacesInMultiwordNodes(constituentTree);
		}
		removeAnnotation(constituentTree, "MWU");
		flattenUnaryChains(constituentTree);
		
		// Adding nodes/branches
//		addPhrasalNodes(constituentTree);
		
		// Empty categories
//		System.out.println("\nIntermediate tree:\n" + constituentTree);
		addNullSJBranch(constituentTree);
		addVerbEllipsisBranch(constituentTree);
		addEllipsisBranch(constituentTree);
		addGapBranch(constituentTree);
		flattenUnaryChains(constituentTree);
//		System.out.println("\nIntermediate tree:\n" + constituentTree);

		// Uncomment to skip the remaining code
//		if (true) {
//			return null;
//		}


		// Temporary fixes, hacks, etc. for v2 and v3
		addXbarTag(constituentTree);
		fix_removeXbarOverX(constituentTree);
		fix_sns2s(constituentTree);
		fix_renameNodes(constituentTree);
		fix_addLexicalPreterminalTags(constituentTree);
		fix_removeFunctionFromCoordinatedCPs(constituentTree);
		fix_renameComparatives(constituentTree);
		remove_D_over_CARDorA(constituentTree);
		addSynFuncToCARDs(constituentTree);
		addSynFuncToSComplement(constituentTree);
		fix_Modifier2Specifier(constituentTree);
		fix_removeExtraSynSemTags(constituentTree);
		fix_Gerunds(constituentTree);
		fix_futureAuxiliaryComplement(constituentTree);
		fix_properNamesAsModifiers(constituentTree);
		fix_NbarOverCoordinatedNPs(constituentTree);
		fix_removeSlashesOutsidePath(constituentTree);
		
		fix_markSubjectOfComplexPredicates(constituentTree);
		fix_relabelARGn(constituentTree);
		fix_relabelAnticausative(constituentTree);
		fix_markControl(constituentTree);
		flattenUnaryChains(constituentTree);
		addXbarTag(constituentTree);
		
		return constituentTree.toString();
	}
	
	
	/*
	 * Operations over String
	 */
	
	
	/**
	 * Creates a Tree structure from the String representation of a tree.
	 * <p>
	 * This method assumes that the String is already in a format that can be read by the Tregex/TSurgeon Stanford API.<br>
	 * See <code>renameParenthesisInLeaves</code> and <code>preprocessTreeString</code> for the String pre-processing methods.
	 * 
	 * @param treeStr  the tree, represented as String (in a format suitable for the API)
	 * @return         the tree, represented using the Tree structure
	 */
	private Tree string2tree(String treeStr) {
		Tree tree = null;
		try {
			TreeReader reader = new PennTreeReader(new StringReader(treeStr), treeFactory);
			tree = reader.readTree();
		} catch (java.io.IOException e) {
			System.err.println("Error creating phrase structure tree: " + e.getMessage());
			System.err.println("The input String was: " + treeStr);
		}
		return tree;
	}
	

	/**
	 * Replaces parenthesis in leaves by unambiguous symbols.
	 * <p>
	 * The leaves may contain parenthesis that will be mistaken as structure.
	 * This method replaces those parenthesis by unambiguous symbols, following the Penn Treebank style:
	 * <ul>
	 * <li>'(' is replaced by '-lrb-'
	 * <li>')' is replaced by '-rrb-'  
	 * </ul>
	 * Requires the derivation tree since only that tree allows us to unambiguously determine if the parenthesis are part of the tree or part of the structure.<br>
	 * After using the derivation tree to find leaves containing parenthesis, and replacing those parenthesis with unambiguous symbols, it repeats the same replacement in the constituency tree.
	 * <p>
	 * TODO: Does it work with very long leaves? (leaves in constituency tree are limited to 30 characters)
	 * 
	 * @param  psTreeStr     the String representation of the constituency tree
	 * @param  derivTreeStr  the String representation of the derivation tree
	 * @return               the String representation of the constituency tree, with parenthesis in leaves replaced
	 */
	private String renameParenthesisInLeaves(String psTreeStr, String derivTreeStr) {
		Matcher m = quotedString.matcher(derivTreeStr);
		while (m.find()) {
			String leaf = m.group(1);
			if (leaf.contains("(") | leaf.contains(")")) {
				String newLeaf = leaf.replace("(", "-lrb-").replace(")", "-rrb-");
				psTreeStr = psTreeStr.replace(leaf, newLeaf);
			}
		}
		return psTreeStr; 
	}
	
	/**
	 * Pre-processes the String representing the [constituency|derivation] tree into a format suitable for reading by the Stanford Tree API being used.
	 * <p>
	 * The pre-processing consists of the following steps:
	 * <ol>
	 * <li> Making leaves lowercase
	 * <li> Removing unecessary information (only for derivation tree)
	 * <li> Fixing nodes whose name starts with a number (only for derivation tree)
	 * <li> Fixing nodes with "?" category (only for constituency tree)
	 * <li> Removing parenthesis that enclose leaves
	 * </ol>
	 *   
	 * @param treeStr  the tree, represented as String (in the Redwoods export format)
	 * @return         the tree, represented as String (in a format suitable for the API)
	 */
	private String preprocessTreeString(String treeStr) {
		treeStr = leavesToLowerCase(treeStr);
		if (treeStr.matches("^\\(\\d+.*")) {
			treeStr = preprocessDerivTreeString(treeStr);
		} else {
			treeStr = treeStr.replaceAll("\\(\\?", "(X_?");
		}
		treeStr = removeLeafParenthesis(treeStr);
		return treeStr;		
	}
	
	
	/**
	 * Converts all leaves to lowercase.
	 * <p>
	 * A leaf is defined as any sequence between enclosing parenthesis that itself does not contain parenthesis.
	 *  
	 * @param  treeStr the String representation of the tree
	 * @return         the String representation of the tree, with all leaves in lowercase
	 */
	private String leavesToLowerCase(String treeStr) {
		Pattern p = Pattern.compile("\\(([^()]+)\\)");
		Matcher m = p.matcher(treeStr);
		StringBuffer sb = new StringBuffer();
		while (m.find()) {
			m.appendReplacement(sb, m.group().toLowerCase());
		}
		m.appendTail(sb);
		return sb.toString();
	}
	
	
	/**
	 * Removes unecessary information from the derivation trees in the Redwoods export.
	 * <p>
	 * The derivation trees in the Redwoods export include information such as chart parser rule IDs, rule probabilities and surface form indexes.<br>
	 * This method removes that information from the String representation of the tree, leaving only the rule names and the leaves. 
	 * <p>
	 * For instance:<br>
	 * (0 RULE 0 0 0 ("leaf" 0 0)) -> (RULE (leaf))
	 * <p>
	 * This method keeps the parenthesis that enclose the leaf. These are removed later on by the <code>removeLeafParenthesis</code> method.
	 * <p>
	 * This method adds a dummy prefix to rule names that start with a number, since <code>removeLeafParenthesis</code> assumes that node labels start with a capital letter.
	 * 
	 * @param  derivTreeString  the String representation of the derivation tree
	 * @return                  the String representation of the derivation tree, with the extra information removed
	 */
	private String preprocessDerivTreeString(String derivTreeString) {
		String result = derivTreeString;
		result = result.replaceAll("\\(\\d+\\s+", "("); // Remove chart parser id
		result = result.replaceAll("\\((\\d+)", "(X_$1"); // Add a dummy prefix to rule names that start with a number
		result = result.replaceAll("\\((\\S+)\\s+\\S+\\s+\\d+\\s+\\d+\\s+", "($1 "); // Remove rule probability and constituent start/end index
		result = result.replaceAll("\\s+\\d+\\s+\\d+\\)", ")"); // Remove leaf start/end index
		result = result.replaceAll("\\(\"", "(").replaceAll("\"\\)", ")"); // Remove leaf quotation marks
		return result;
	}


	/**
	 * Removes parenthesis enclosing leaves from the String representation of the tree.
	 * <p>
	 * The phrase structure trees in the Redwoods export have the leaf nodes within parenthesis, which causes the Stanford Tregex API to create spurious nodes.<br> 
	 * This method takes a String representing a phrase structure tree and returns a String where the parenthesis that enclose leaves have been removed.
	 * <p>
	 * For instance:<br>
	 * (N (gato)) -> (N gato)
	 * <p>
	 * This method assumes that leaf nodes start with lowercase letters and that non-leaf nodes start with uppercase letters.
	 * Accordingly, <code>leavesToLowerCase</code> must be previously run over the String.
	 * <p>
	 * This assumption is needed to properly handle multi-word leafs. For instance:<br>
	 * (Node (leaf1 leaf2)) -> (Node leaf1 leaf2), and not (Node (leaf1 leaf2))  
	 * 
	 * @param  treeStr  the String representation of the tree
	 * @return          the String representation of the tree, with leaf parenthesis removed
	 */
	private String removeLeafParenthesis(String treeStr) {
		String result = treeStr;
		Pattern p = Pattern.compile("\\(([^\\p{javaUpperCase}][^()]*)\\)");
		Matcher m = p.matcher(result);
		while (m.find()) {
			result = m.replaceFirst(m.group(1));
			m = p.matcher(result);
		}
		return result;
	}

	
	/**
	 * Expands a contracted sequence.
	 * <p>
	 * This is a helper method used by <code>fixClosedClassMWUNodes</code>.
	 * <p>
	 * TODO: Use a HashMap instead?
	 * TODO: aonde,
	 *       àquel[ae]s?,
	 *       n(aquel[ae]s?|algu(?:m|as?|ns))
	 *       d(aqui|ali|acolá|aí|além|aquém|aquilo|entre|isso|isto|onde|outr[ao]s?|aquel[ae]s?|el[ae]s?|ess[ae]s?|algu(?:m|as?|ns)|u(?:m|as?|ns)), 
	 *       comigo, contigo, connosco, convosco,
	 *       among others... cf. (Silva, 2007)
	 * TODO: ambiguous contractions: consigo, desse, desses, deste, destes, na, nele, no, nos, pela, pelas, pelo 
	 * 
	 * @param result
	 * @return
	 */
	private String expandContractionInMWU(String mwu) {
		String result = mwu;
		
		// a + X : à, às, ao, aos
		result = result.replaceAll("\\bà(s?)\\b", "a_ a$1");
		result = result.replaceAll("\\bao(s?)\\b", "a_ o$1");
		
		// de + X : desta, destas, deste, destes
		result = result.replaceAll("\\bdest([ae])(s?)\\b", "de_ est$1$2");
		// de + X : da, das, do, dos
		result = result.replaceAll("\\bd([ao])(s?)\\b", "de_ $1$2");
		
		// em + X : na, nas, no, nos
		result = result.replaceAll("\\bn([ao])(s?)\\b", "em_ $1$2");
		// em + X : num
		result = result.replaceAll("\\bnum(a?)\\b", "em_ um$1");
		
		// por + X : pela, pelas, pelo, pelos
		result = result.replaceAll("\\bpel([ao])(s?)\\b", "por_ $1$2");
		
		return result;
	}
	
	
	/**
	 * Miscellaneous fixes to the annotated sentence.
	 * <ul>
	 * <li>Remove spacing aroung decimal separator.
	 * <li>Move period inside named-entity.
	 * </ul>
	 * @param annotated
	 */
	private String fixBadTokenizationInSentence(String annotated) {
//		System.out.println("\nFix bad tokenization in sentence...");
		annotated = annotated.replaceAll("(\\d+)/DGT\\[O] ,/PNT\\[O] (\\d+)/DGT\\[O]", "$1,$2/DGT[O]");
		annotated = annotated.replaceAll("(\\w+)/(\\w+\\[[BI]-\\w+]) \\./PNT\\[O]", "$1./$2");
		return annotated;
	}

	
	/*
	 * Operations over Tree
	 */


	/**
	 * Checks if two trees have the same structure. That is, if they are equal
	 * apart from differences in node labels.
	 * <p>
	 * This method works by comparing the node numbers of all leaves.
	 * <p>
	 * NB: While it is true that two trees with the same structure must have
	 * matching node numbers in their leaves, the reverse implication is not
	 * necessarily true. That is, it is possible for two trees to have matching
	 * node numbers in their leaves but non-matching structures. In such cases,
	 * this method will return the wrong answer.
	 * 
	 * @param t1
	 *            one tree
	 * @param t2
	 *            the other tree
	 * @return whether the trees have a matching structure
	 */
	private boolean sameStructure(Tree t1, Tree t2) {
		List<Integer> nnT1 = new ArrayList<Integer>();
		for (Tree l : t1.getLeaves()) {
			nnT1.add(l.nodeNumber(t1));
		}
		List<Integer> nnT2 = new ArrayList<Integer>();
		for (Tree l : t2.getLeaves()) {
			nnT2.add(l.nodeNumber(t2));
		}
		return nnT1.equals(nnT2);
	}
	

	/**
	 * Sanity check that compares nº of leaves in the tree with the nº of tokens in the annotated sentence.
	 * 
	 * @param tree
	 * @param annotated
	 * @return
	 */
	private boolean sameLength(Tree tree, String annotated) {
		int lengthTree = tree.getLeaves().size();
		int lengthSent = annotated.split(" ").length;
		return (lengthTree == lengthSent);
	}
	
	
	/*
	 * Fixes 
	 */
	

	/**
	 * This is a fix for a problem in v3: 
	 * The grammar does not assign a category to "%"/"por cento".
	 * <p> 
	 * NB: Whatever the "percent" tag is, it must be added to <code>addXbarTag</code>.
	 * 
	 * @param tree
	 */
	private void fix_PercentNodes(Tree tree) {
//		System.out.println("\nFix percent nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/X_\\?/ < /%|(?:por cento)|PERCENTP?/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				node.label().setFromString(node.isPreTerminal() ? "PERCENT" : "PERCENTP");
				matcher.reset();
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * This is a fix for a problem in v3: 
	 * The grammar does not assign a category to the "e" in "entre ... e".
	 * <p> 
	 * NB: Whatever the "e" tag is, it must be added to <code>addXbarTag</code>.
	 * 
	 * @param tree
	 */
	private void fix_EntreENodes(Tree tree) {
//		System.out.println("\nFix \\"entre...e\\" nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/X_\\?/ < /e|CONJP?/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				if (node.isPreTerminal()) {
					node.label().setFromString("CONJ");
				} else {
					node.label().setFromString("CONJP");
				}
				matcher.reset();
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	

	/**
	 * Removes slash markings from nodes that are not in the path(s) connecting the filler with its gap(s).
	 * 
	 * @param tree
	 */
	private void fix_removeSlashesOutsidePath(Tree tree) {
//		System.out.println("\nRemove slashes outside gap-filler paths...");
		
		try {
			TregexPattern pathTregex = TregexPattern.compile("/\\//=top [ <+(/\\//) (/\\// < (__=gap <: /\\*(?:GAP|TOUGH)\\*/)) | < (__=gap <: /\\*(?:GAP|TOUGH)\\*/)] !> /\\//");
			TregexMatcher pathMatcher = pathTregex.matcher(tree);
			HashMap<Integer, List<Integer>> paths = new HashMap<Integer, List<Integer>>();
			
			// Collect endpoints of all gap-filler paths
			// NB: Storing the Tree directly would be better, but it doesn't seem to work. Coordinated gap nodes are not seen as being different.
			while (pathMatcher.find()) {
				Tree topSlash = pathMatcher.getNode("top");
				Tree gapNode = pathMatcher.getNode("gap");
				int topEndpoint = topSlash.nodeNumber(tree);
				int gapEndpoint = gapNode.nodeNumber(tree);
				if (!paths.containsKey(topEndpoint)) {
					paths.put(topEndpoint, new ArrayList<Integer>());
				}
				if (!paths.get(topEndpoint).contains(gapEndpoint)) {
					paths.get(topEndpoint).add(gapEndpoint);
				}
			}
				
			// Slash removal
			TregexPattern slashPattern = TregexPattern.compile("/\\// <+(/\\//) (/\\//=slash < __)");
			for (int topEndpoint : paths.keySet()) {
				Tree topSlash = tree.getNodeNumber(topEndpoint);
				TregexMatcher slashMatcher = slashPattern.matcher(topSlash);
				while (slashMatcher.matches()) {
					Tree candidateSlash = slashMatcher.getNode("slash");
					boolean dominatesGap = false;
					for (int gapEndpoint : paths.get(topEndpoint)) {
						Tree gapNode = tree.getNodeNumber(gapEndpoint);
						dominatesGap = dominatesGap || candidateSlash.dominates(gapNode);
					}
					if (!dominatesGap) {
						candidateSlash.label().setFromString(candidateSlash.label().toString().replaceFirst("/[^*]+", ""));
					}
				}
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Helper method...
	 *  
	 * @param tree
	 * @param from
	 * @param to
	 */
	private void aux_relabelSlashes(Tree tree, String from, String to) {
//		System.out.println("\nRelabel connected slashes...");
		try {
			String fromPattern = "/\\/" + from + "/";
			String fullPattern = fromPattern + " <+(" + fromPattern + ") " + fromPattern + "=n";
//			System.out.println("Pattern: " + fullPattern);
			TregexPattern slashTregex = TregexPattern.compile(fullPattern);
			TregexMatcher slashMatcher = slashTregex.matcher(tree);
			List<Tree> slashList = new ArrayList<Tree>();
			slashList.add(tree);
			while (slashMatcher.matches()) {
				slashList.add(slashMatcher.getNode("n"));
			}
			for (Tree node : slashList) {
				node.label().setFromString(node.label().toString().replaceFirst("/" + from, "/" + to));
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	

	
	/**
	 * 
	 * @param tree
	 * @return
	 */
	private void fix_relabelARGn(Tree tree) {
//		System.out.println("\nRelabel ARGn tags...");
		try {
			TregexPattern argaTregex = TregexPattern.compile("/^NP-[^-]+-ARGA/=arga $ /^VP\\/?/=vp");
			TregexMatcher argaMatcher = argaTregex.matcher(tree);
			
			while (argaMatcher.find()) {
				Tree argaNode = argaMatcher.getNode("arga");
				Tree siblingVPNode = argaMatcher.getNode("vp");
//				boolean isComplexPredicate = argaNode.label().toString().endsWith("cp");
				
				String oldARGALabel = argaNode.label().toString();
				String newARGALabel = oldARGALabel.replaceFirst("ARGA", "ARG1");
				argaNode.label().setFromString(newARGALabel);
				
				// Propagate the change to co-indexed nodes and slashes
				if (oldARGALabel.matches(".+_\\d+$")) {
					// Relabel any co-indexed nodes
					TregexPattern relabelTregex = TregexPattern.compile(oldARGALabel + "=n");
					TsurgeonPattern relabelTsurgeon = Tsurgeon.parseOperation("relabel n " + newARGALabel);
					Tsurgeon.processPattern(relabelTregex, relabelTsurgeon, tree);
					// Find the top slash and relabel all slashes connected to it
					boolean finished = false;
					Tree topSlash = siblingVPNode;
					while (!finished) {
						Tree parent = topSlash.parent(tree);
						finished = parent == null || !parent.label().toString().contains("/"); 
						if (!finished) {
							topSlash = parent;
						}
					}
					String oldSlash = topSlash.label().toString().replaceFirst("[^/]+/", "");
					String newSlash = oldSlash.replaceFirst("ARGA", "ARG1");
					aux_relabelSlashes(topSlash, oldSlash, newSlash);
				}

				// Update relevant ARGn
				TregexPattern argnFinderTregex = TregexPattern.compile("/^VP\\/?/ <+(/^(?:(:?V'?|VP(?:-C)?)(?:\\/.+)?|PP.*)$/) (/^V\\/?/ $ /P-[^-]+-ARG\\d+$/=arg)");
				TregexMatcher argnFinderMatcher = argnFinderTregex.matcher(siblingVPNode);
				while (argnFinderMatcher.matches()) {
					Tree argnNode = argnFinderMatcher.getNode("arg");
					String oldLabel = argnNode.label().toString();
					String newLabel = null;
					if (oldLabel.contains("ARG1")) {
						newLabel = oldLabel.replaceFirst("ARG1", "ARG2");
					} else {
						newLabel = oldLabel.replaceFirst("ARG2", "ARG3");
					}
					argnNode.label().setFromString(newLabel);
				}
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 */
	private void fix_relabelAnticausative(Tree tree) {
//		System.out.println("\nRelabel anticausative...");
		
		TregexPattern tregex = null;
		TregexMatcher matcher = null;
		try {
			
			// Nudge the mark to the correct place
			tregex = TregexPattern.compile("/VP\\*AC\\*/=c $ PNT > VP=t");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree childNode = matcher.getNode("c");
				Tree topNode = matcher.getNode("t");
				childNode.label().setFromString(childNode.label().toString().replaceFirst("\\*AC\\*", ""));
				topNode.label().setFromString(topNode.label().toString() + "*AC*");
				matcher.reset();
			}
			
			// Annotate sibling of marked node
			tregex = TregexPattern.compile("/\\*AC\\*/=vp $ /NP-SJ-ARG1/=sj");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree vpNode = matcher.getNode("vp");
				Tree sjNode = matcher.getNode("sj");
				vpNode.label().setFromString(vpNode.label().toString().replaceFirst("\\*AC\\*", ""));
				if (sjNode.label().toString().matches(".+_\\d+")) {
					String oldSJLabel = sjNode.label().toString();
					String newSJLabel = oldSJLabel.replaceFirst("ARG1", "ARG2ac");
					// Relabel any co-indexed nodes
					TregexPattern relabelTregex = TregexPattern.compile(oldSJLabel + "=n");
					TsurgeonPattern relabelTsurgeon = Tsurgeon.parseOperation("relabel n " + newSJLabel);
					Tsurgeon.processPattern(relabelTregex, relabelTsurgeon, tree);
					// Find the top slash and relabel all slashes connected to it
					boolean finished = false;
					Tree topSlash = vpNode;
					while (!finished) {
						Tree parent = topSlash.parent(tree);
						finished = parent == null || !parent.label().toString().contains("/"); 
						if (!finished) {
							topSlash = parent;
						}
					}
					String oldSlash = topSlash.label().toString().replaceFirst("[^/]+/", "");
					String newSlash = oldSlash.replaceFirst("ARG1", "ARG2ac");
					aux_relabelSlashes(topSlash, oldSlash, newSlash);
				} else {
					sjNode.label().setFromString(sjNode.label().toString().replaceFirst("ARG1", "ARG2ac"));
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	

	/**
	 * 
	 * @param tree
	 */
	private void fix_markSubjectOfComplexPredicates(Tree tree) {
//		System.out.println("\nMark subject of complex predicates...");
		TregexPattern tregex = null;
		TregexMatcher matcher = null;
		try {
			
			// Nudge the mark to the correct place
			tregex = TregexPattern.compile("/\\*CP\\*/=b !$ @NP > VP=t");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree bNode = matcher.getNode("b");
				Tree tNode = matcher.getNode("t");
				bNode.label().setFromString(bNode.label().toString().replaceFirst("\\*CP\\*", ""));
				tNode.label().setFromString(tNode.label().toString() + "*CP*");
				matcher.reset();
			}
			
			// Annotate sibling of marked node
			tregex = TregexPattern.compile("/\\*CP\\*/=vp $ /NP-SJ-ARG./=sj");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree vpNode = matcher.getNode("vp");
				Tree sjNode = matcher.getNode("sj");
				vpNode.label().setFromString(vpNode.label().toString().replaceFirst("\\*CP\\*", ""));
				String oldLabel = sjNode.label().toString();
				String newLabel = oldLabel.replaceFirst("(_\\d+)?$", "cp$1");
				sjNode.label().setFromString(newLabel);
				
				// Propagate the change to co-indexed nodes and slashes
				if (newLabel.matches(".+cp_\\d+$")) {
					// Relabel any co-indexed nodes
					TregexPattern relabelTregex = TregexPattern.compile(oldLabel + "=n");
					TsurgeonPattern relabelTsurgeon = Tsurgeon.parseOperation("relabel n " + newLabel);
					Tsurgeon.processPattern(relabelTregex, relabelTsurgeon, tree);
					// Find the top slash and relabel all slashes connected to it
					boolean finished = false;
					Tree topSlash = vpNode;
					while (!finished) {
						Tree parent = topSlash.parent(tree);
						finished = parent == null || !parent.label().toString().contains("/"); 
						if (!finished) {
							topSlash = parent;
						}
					}
					String oldSlash = topSlash.label().toString().replaceFirst("[^/]+/", "");
					String newSlash = oldSlash + "cp";
					aux_relabelSlashes(topSlash, oldSlash, newSlash);
				}
				
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_markControl(Tree tree) {
//		System.out.println("\nMark control...");
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*CONTROL\\*/ $ @NP=np");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.findNextMatchingNode()) {
				Tree cNode = matcher.getMatch();
				Tree npNode = matcher.getNode("np");
				cNode.label().setFromString(cNode.label().toString().replaceFirst("\\*CONTROL\\*", ""));
				npNode.label().setFromString(npNode.label().toString().replaceFirst("(ARG\\d)", "$1"+ "1"));
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	

	/**
	 * 
	 * @param tree
	 */
	private void fix_NbarOverCoordinatedNPs(Tree tree) {
//		System.out.println("\nFix N' over coordinated NPs...");
		try {
			TregexPattern tregex = TregexPattern.compile("N' <1 NP <2 NP");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.findNextMatchingNode()) {
				Tree node = matcher.getMatch();
				boolean finished = false;
				while (!finished) {
					node.label().setFromString("NP");
					node = node.parent(tree);
					finished = node == null || !node.label().toString().contentEquals("N'");
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_properNamesAsModifiers(Tree tree) {
//		System.out.println("\nFix proper names modifying nouns...");
		try {
			TregexPattern tregex = TregexPattern.compile("N=n < /PNM:/ $ (N|N' !< CONJ)");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("relabel n N-M-M");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_futureAuxiliaryComplement(Tree tree) {
//		System.out.println("\nFix complement of future auxiliary...");
		try {
			TregexPattern tregex = TregexPattern.compile("VP-C-ARG1=c $ (/^(?:VP|V|V')/ << /:IR:/)");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("relabel c VP-C");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * This is a fix for problems in v3 related with gerunds
	 * 
	 * @param tree
	 */
	private void fix_Gerunds(Tree tree) {
//		System.out.println("\nFix gerunds...");
		try {
			TregexPattern tregex = TregexPattern.compile("/[aei]ndo(?:$|\\/GER|\\/V:.+:GER:)/ >: @V [ >> (/(?:VP|V'|V)-M/=x !< PNT) | >> (/(?:VP|V'|V)-M/ < VP=x < PNT=p) ]");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.findNextMatchingNode()) {
				Tree xNode = matcher.getNode("x");
				Tree pNode = matcher.getNode("p");
//				System.out.println("Found: " + matcher.getMatch());
//				System.out.println("Target: " + xNode);
//				System.out.println("P?: " + pNode);
				String gramsemTags = xNode.label().toString().replaceFirst("^[^-]+", "");
				xNode.label().setFromString(xNode.label().toString().replaceFirst("-[^/]+", ""));
				Tree newBranch = string2tree("(S" + gramsemTags + " (NP-SJ-ARG1 *NULL*))");
				newBranch.addChild(xNode);
//				System.out.println("Branch: " + newBranch);
				Tree parent = xNode.parent(tree);
				parent.setChild(parent.indexOf(xNode), newBranch);
				if (pNode != null) {
					parent.label().setFromString(parent.label().toString().replaceFirst("VP", "S"));
				}
//				System.out.println("Final: " + parent);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * This is a fix for problems in v3 related with participles
	 * 
	 * @param psTree
	 */
	private void fix_Participles(Tree psTree, Tree derivTree) {
//		System.out.println("\nFix participles...");

		try {
			TregexPattern tregex = null;
			TregexMatcher matcher = null;
			TsurgeonPattern tsurgeon = null;

			// Bug: For participles not in the lexicon, the grammar assigns a single pre-terminal A instead of (V (A ...))
			// Fix: Insert a pre-terminal V
			// TODO: This fix changes the tree structure. If we manage to avoid this (and relabel instead), this method can be moved together with the other fixes.
			tregex = TregexPattern.compile("/GENERIC-PASSIVE-PARTICIPLE/");
			matcher = tregex.matcher(derivTree);
			List<Tree> nodesPassive = new ArrayList<Tree>();
			while (matcher.find()) {
				Tree derivNode = matcher.getMatch();
				int nodeNumber = derivNode.nodeNumber(derivTree);
				Tree psNode = psTree.getNodeNumber(nodeNumber);
				if (!psNode.label().toString().startsWith("V")) {
					nodesPassive.add(psNode);
				}
			}
			for (Tree passiveNode : nodesPassive) {
				String newLabel = "V";
				// Special case to handle nodes marked with a lexical type
				if (passiveNode.label().toString().contains("*lextype:")) {
					newLabel = "V" + passiveNode.label().toString().replaceFirst(".*(\\*lextype:[^*]+\\*).*", "$1");
					passiveNode.label().setFromString(passiveNode.label().toString().replaceFirst("\\*lextype:[^*]+\\*", ""));
				}
				Tree[] newNode = {treeFactory.newTreeNode(newLabel, passiveNode.getChildrenAsList())};
				passiveNode.setChildren(newNode);
			}
			
			// Bug: V participles project A (and should project AP)
			// Fix: Relabel
			tregex = TregexPattern.compile("@VP [!< @V & !< @V' & !< @VP & < (A <<: /^V(?:\\*lextype:.+\\*)?$/)]");
			matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				boolean finished = false;
				while (!finished) {
					node.label().setFromString("AP");
					Tree parent = node.parent(psTree);
					if (parent != null && parent.label().toString().startsWith("VP")) {
						node = parent;
					} else {
						finished = true;
					}
				}
				matcher.reset();
			}
			
			// Bug: V participles project A (and should project AP)
			// Fix: Relabel
			tregex = TregexPattern.compile("@AP << (A=a <: /^V(?:\\*lextype:.+\\*)?$/)");
			matcher = tregex.matcher(psTree);
			tsurgeon = Tsurgeon.parseOperation("relabel a V"); // "excise a a" fails for some reason
			Tsurgeon.processPattern(tregex, tsurgeon, psTree);
			
			// Bug: V participles project A (and should project AP)
			// Fix: Relabel
			tregex = TregexPattern.compile("@A=a <<: /^V(?:-.+)?(?:\\*lextype:.+\\*)?$/=v");
			matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree aNode = matcher.getNode("a");
				Tree vNode = matcher.getNode("v");
				aNode.label().setFromString(aNode.label().toString().replaceFirst("A", "AP"));
				// Special case to handle nodes marked with a lexical type
				if (vNode.label().toString().contains("*lextype:")) {
					String mark = vNode.label().toString().replaceFirst(".*(\\*lextype:[^*]+\\*).*", "$1");
					vNode.label().setFromString("V" + mark);
				} else {
					vNode.label().setFromString("V");
				}
			}
			
			// (VP (A (A ... (A planeada)))
			// eCTMP/000109/35620
			tregex = TregexPattern.compile("VP=v < (A=a <<: (/^A(?:\\*lextype:.+\\*)?$/ <: (__ !< __)))");
			matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree vNode = matcher.getNode("v");
				vNode.label().setFromString("AP");
				Tree aNode = matcher.getNode("a");
				while (!aNode.isLeaf()) {
					// Special case to handle nodes marked with a lexical type
					if (aNode.label().toString().contains("*lextype:")) {
						String mark = aNode.label().toString().replaceFirst(".*(\\*lextype:[^*]+\\*).*", "$1");
						aNode.label().setFromString("V" + mark);
					} else {
						aNode.label().setFromString("V");
					}
					aNode = aNode.firstChild();
				}
			}
			
			// (AP-PRD (V ...) (VP-PRD (CONJ e) (V ...)))
			// eCTMP/000533/69023
			tregex = TregexPattern.compile("@AP <1 (V <<: /calcinados/) <2 (@VP=vp <1 /^CONJ/ <2 (V <<: /mutilados/))");
			matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree vpNode = matcher.getNode("vp");
				vpNode.label().setFromString("AP");
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}
		
	}


	/**
	 * Replaces the leaf nodes in the constituent tree with the corresponding leaf nodes from the derivation tree.
	 * <p>
	 * The Redwoods export of the constituent tree has a builtin limit of 30 characters for leaves, which causes some leaves in the constituent tree to be cut.
	 * Since this limit does not exist in the Redwoods export of the derivation tree, we substitute every leaf in the constituent tree by the corresponding leaf in the derivation tree.
	 * 
	 * @param psTree     the phrase structure tree
	 * @param derivTree  the derivation tree
	 */
	private void replaceLeavesInConstituentTree(Tree psTree, Tree derivTree) {
//		System.out.println("\nFix leaves in constituent tree...");
		List<Tree> psTreeLeaves = psTree.getLeaves();
		List<Tree> derivTreeLeaves = derivTree.getLeaves();
		for (int i = 0; i < psTreeLeaves.size(); i++) {
			psTreeLeaves.get(i).label().setFromString(derivTreeLeaves.get(i).label().toString());
		}
	}


	/**
	 * Expands contractions in closed class MWU leaves.
	 * <p>
	 * Contractions in closed class MWU (e.g. "do que", "à risca") are not expanded in the tsdb exported trees.
	 * However, since they are expanded in the annotated sentences, this lead to a misalignment between tokens and leaves.
	 * <p>
	 * To find the contracted sequences, this method uses the rule name in the derivation tree.
	 * The expansion itself if handled by the <code>expandContractionInMWU</code> helper method. 
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void expandContractionsInClosedClassMWUNodes(Tree psTree, Tree derivTree) {
//		System.out.println("\nFix closed class MWU leaves in constituent tree...");
		try {
			TregexPattern tregex = TregexPattern.compile("/GENERIC-(?:ADVERB|PREPOSITION)-ENTRY|\\+/");
			TregexMatcher matcher = tregex.matcher(derivTree);
			while (matcher.find()) {
				Tree derivNode = matcher.getMatch();
//				System.out.println("Found in derivation tree: " + derivNode);
				if (derivNode.isPreTerminal()) {
					Tree leafNode = psTree.getNodeNumber(derivNode.nodeNumber(derivTree)).getChild(0);
//					System.out.println("Corresponding leaf: " + leafNode);
					leafNode.label().setFromString(expandContractionInMWU(leafNode.toString()));
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/*
	 * Methods that decorate the tree with information (usually on empty categories) to be used later. 
	 */
	
	
	/**
	 * Marks modifier nodes so that they can still be detected even if syntactic function tags are removed.
	 * <p>
	 * Namely, <code>preprocessParentheticalModifierPunctuation</code> must know where modifiers are.
	 * <p>
	 * NB: This is not strictly needed anymore, since syntactic function tags are now kept throughout the whole process.
	 * <p>
	 * TODO: Externalize the label and the tag used to indicate that the constituent is a modifier.
	 */
	private void annotateModifierNodes(Tree tree) {
//		System.out.println("\nAnnotate modifier nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^[^-\\/]+-M/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.findNextMatchingNode()) {
				Tree modNode = matcher.getMatch();
				if (!modNode.isLeaf()) {
					modNode.label().setFromString(modNode.label().toString() + "*M*");
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}

	
	/**
	 * Marks ellipsis nodes with information on whether the elided constituent is the left child or the right child of the node.
	 * <p>
	 * To know if a node is missing an elided constituent, and whether it is a left or right child, this method must look at the derivation tree.
	 * The corresponding rule name indicates that there is a missing constituent as well as whether that constituent should occur as the left child or as the right child of the node.
	 * <p>
	 * NB: To ensure that there is structural alignment between the constituent tree and the derivation tree, this method must be run before performing any structural modifications to the constituent tree.
	 * <p>
	 * TODO: Externalize the labels used to indicate left/right and the rule name that indicates a missing constituent.
	 * 
	 * @param psTree     the phrase structure tree
	 * @param derivTree  the derivation tree
	 */
	private void annotateEllipsisNodes(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate ellipsis nodes with left/right markings...");
		try {
			
			// Noun ellipsis
			TregexPattern nounTregex = TregexPattern.compile("/^MISSING-NOUN.*/ ?<< QUE_COMPARATIVE=q");
			TregexMatcher nounMatcher = nounTregex.matcher(derivTree);
			while (nounMatcher.find()) {
				Tree derivNode = nounMatcher.getMatch();
				String ellipsisLabel = (derivNode.label().toString().contains("-FINAL-") && nounMatcher.getNode("q") == null) ? "*RIGHT*" : "*LEFT*";
				Tree psNode = psTree.getNodeNumber(derivNode.nodeNumber(derivTree));
				psNode.label().setFromString(psNode.label().toString() + ellipsisLabel);
			}
			
			// Verb ellipsis
			TregexPattern verbTregex = TregexPattern.compile("/^MISSING-COPULA/");
			TregexMatcher verbMatcher = verbTregex.matcher(derivTree);
			while (verbMatcher.find()) {
				Tree derivNode = verbMatcher.getMatch();
				Tree psNode = psTree.getNodeNumber(derivNode.nodeNumber(derivTree));
				psNode.label().setFromString(psNode.label().toString() + "*LEFT*");
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}

	
	/**
	 * Marks null subject nodes with semantic argument information taken from the derivation tree.
	 * <p>
	 * To know the semantic argument of the null subject, this method must look at the derivation tree.
	 * The marking will be used later on when adding a null subject branch to allow labeling that branch with the proper ARGx semantic argument tag.
	 * <p>
	 * The null subject node will be marked as *ARG1*, except in the following cases:
	 * <ul>
	 * <li>Expletive: Does not receive an *ARGx* marking, only ** (i.e. empty).
	 * <li>Passive: Receives an *ARG2* marking.
	 * <li>Causative alternation: Receives an *ARGA* marking.
	 * </ul>
	 * The method that adds the null subject branch will copy the marking and use it as a semantic argument tag.
	 * <p>
	 * TODO: Get the SNS tag from Properties.
	 * TODO: Check special cases: Raising verbs, passive together with alternation (=ARG1?), etc.
	 * TODO: PPA will stop being used in version 3 of the grammar.
	 * 
	 * @param psTree     the phrase structure tree
	 * @param derivTree  the derivation tree
	 */
	private void annotateNullSJNodes(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate null subject nodes...");
		try {
			TregexPattern snsTregex = TregexPattern.compile("/^SNS/");
			TregexMatcher snsMatcher = snsTregex.matcher(psTree);
			while (snsMatcher.find()) {
				Tree snsNode = snsMatcher.getMatch();
//				System.out.println("\nFound: " + snsNode);

				Tree snsDerivNode = derivTree.getNodeNumber(snsNode.nodeNumber(psTree));
				String snsRuleName = snsDerivNode.label().toString();
				String snsMark = null;
				
				// FIXME: Fix for a bug in version 2 of the grammar where NP-FRAGMENT rules are wrongly shown as SNS (they should be S)
				if (snsRuleName.contains("NP-FRAGMENT")) {
//					System.out.println("FIX: Relabel node and skip it (NP-FRAGMENT)");
					snsNode.label().setFromString("S");
					continue;
				}
				
				// Expletive
				if (snsMark == null) {
					if (snsRuleName.contains("EXPLETIVE")) {
						snsMark = "**";
					}
				}
				
				// Causative alternation
				// If the prepreterminal node of the head verb is CAUSATIVE-ALTERNATION (derivation tree)
				if (snsMark == null) {
					TregexPattern searchForCausative = TregexPattern.compile("/^SNS/ <+(/^(?:VP|V'|V|PPA)$/) (V'|V|PPA=v <: (V|PPA <: (__ !< __)))");
					TregexMatcher causativeMatcher = searchForCausative.matcher(snsNode);
					
					while (causativeMatcher.matches()) {
						Tree candidate = causativeMatcher.getNode("v");
						String derivationRule = snsDerivNode.getNodeNumber(candidate.nodeNumber(snsNode)).label().toString();
//						System.out.println("Check candidate for causative alternation: " + candidate + " -> " + verbRuleName);
						if (derivationRule.contains("CAUSATIVE-ALTERNATION")) {
							snsMark = "*ARGA*";
//							candidate.label().setFromString(candidate.label().toString() + "*ALT*"); // Debug marking
						}
					}
				}
				
				// Passive
				if (snsMark == null) {
//					System.out.println("Check passive:\n" + snsNode);
					TregexPattern new_searchForPassive = TregexPattern.compile("/^SNS/ <: (VP <1 V|V' <2 (VP-PRD <+(V|V'|VP) V|V'=v))");
					TregexMatcher new_passiveMatcher = new_searchForPassive.matcher(snsNode);
					while (snsMark == null && new_passiveMatcher.matches()) {
//						System.out.println("Candidate for passive: " + new_passiveMatcher.getMatch());
						Tree vNode = new_passiveMatcher.getNode("v");
						Tree derivNode = snsDerivNode.getNodeNumber(vNode.nodeNumber(snsNode));
//						System.out.println("V is " + derivNode.label().toString());
						if (derivNode.label().toString().contains("PASSIVE-PARTICIPLE")) {
							snsMark = "*ARG2*";
						}
					}
					
					/*
					// PPA only works in v2
					TregexPattern searchForPassive = TregexPattern.compile("/^SNS/ <+(/^(V|PPA)/) (/^(V|PPA)/=v < (/^V/ $++ /^PPA/))");
					TregexMatcher passiveMatcher = searchForPassive.matcher(snsNode);
					while (passiveMatcher.matches()) {
						snsMark = "*ARG2*";
//						Tree candidate = passiveMatcher.getNode("v");
//						candidate.label().setFromString(candidate.label().toString() + "*PASS*"); // Debug marking
					}
					*/
					
				}

				// The default case
				if (snsMark == null) {
					snsMark = "*ARG1*";
				}
				
//				System.out.println("Tag the SNS node with " + snsMark);
				snsNode.label().setFromString(snsNode.label().toString() + snsMark);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
		
	}
	
	
	/**
	 * Marks the slashed node from where the gapped constituent has been extracted with information on whether the gap is the left child or the right child of the node.
	 * <p>
	 * To find the extraction point, and whether the gap is the left or right child, this method must look at the derivation tree.
	 * 
	 * @param psTree     the phrase structure tree
	 * @param derivTree  the derivation tree
	 */
	private void annotateGapNodes(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate gap (slash) nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/EXTRACTED/");
			TregexMatcher matcher = tregex.matcher(derivTree);
			while (matcher.find()) {
				String mark = matcher.getMatch().label().toString().contains("SUBJECT-HEAD_EXTRACTED") ? "*GAPl*" : "*GAPr*";
				Tree psNode = psTree.getNodeNumber(matcher.getMatch().nodeNumber(derivTree));
				psNode.label().setFromString(psNode.label().toString() + mark);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateSubjectRaising(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate subject raising verbs...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^V/ [ >+(/^V/) (/^VP/=n [ > /^S/ $ @NP | >: /^S.*\\/NP.*\\*GAP[lr]\\*/ | >: /^SNS/]) | >+(/^V/) (/^V/=n $ @NP > /V\\+SJ/) ] <: (__ !< __)");
			TregexMatcher matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString().toLowerCase();
				String lexType = entry2type.containsKey(lexEntry) ? entry2type.get(lexEntry) : null;
				boolean isRaising = lexType != null && (lexType.contains("subj_raising"));
				Tree candidateNode = matcher.getNode("n");
//				System.out.println("\nCheck: " + node);
//				System.out.println("Lexicon: " + lexEntry + " -> " + lexType);
				if (isRaising && !candidateNode.label().toString().contains("*CP*")) {
					candidateNode.label().setFromString(candidateNode.label().toString() + "*CP*");
//					System.out.println("Mark: " + candidateNode);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateAuxiliar(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate auxiliar verbs...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^V/ >+(/^V/) (/^V/ > (/^VP/=vp [ > /^S/ $ @NP | >: /^SNS/])) <: (__ !< __)");
			TregexMatcher matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString().toLowerCase();
				String lexType = entry2type.containsKey(lexEntry) ? entry2type.get(lexEntry) : null;
				boolean isAuxiliar = lexType != null && (lexType.contains("verb-compound_tense_aux"));
				Tree candidateNode = matcher.getNode("vp");
//				System.out.println("\nCheck: " + node);
//				System.out.println("Lexicon: " + lexEntry + " -> " + lexType);
				if (isAuxiliar && !candidateNode.label().toString().contains("*CP*")) {
					candidateNode.label().setFromString(candidateNode.label().toString() + "*CP*");
//					System.out.println("Mark: " + candidateNode);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateCopula(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate copula...");
		try {
			// Explicit copula (i.e. present on the tree)
			TregexPattern tregex = TregexPattern.compile("V [ >+(V|V'|VP) (V|V'|VP=x > (VP=n [> /^S/ $ @NP | >: /^SNS/ ] [<2 (@VP !== =x) | <2 (@PP < @VP)])) | >+(V|V') (V|V'=n $ @NP > (/V\\+SJ/ [ $ @VP | $ (@PP < @VP) ])) ] <: (__ !< __)");
			TregexMatcher matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString().toLowerCase();
				String lexType = entry2type.containsKey(lexEntry) ? entry2type.get(lexEntry) : null;
				boolean isRaising = lexType != null && lexType.contains("_copula");
				Tree candidateNode = matcher.getNode("n");
//				System.out.println("\nCheck: " + node);
//				System.out.println("Lexicon: " + lexEntry + " -> " + lexType);
				if (isRaising && !candidateNode.label().toString().contains("*CP*")) {
					candidateNode.label().setFromString(candidateNode.label().toString() + "*CP*");
//					System.out.println("Mark: " + candidateNode);
				}
			}
			
			// Missing copula
			TregexPattern missingCopulaTregex = TregexPattern.compile("/MISSING-COPULA/");
			TregexMatcher missingCopulaMatcher = missingCopulaTregex.matcher(derivTree);
			while (missingCopulaMatcher.find()) {
				Tree derivNode = missingCopulaMatcher.getMatch();
				Tree psNode = psTree.getNodeNumber(derivNode.nodeNumber(derivTree));
				if (psNode.numChildren() == 1 && psNode.getChild(0).label().toString().matches("V-PRD")) {
					psNode.label().setFromString(psNode.label().toString() + "*CP*");
//					System.out.println("Mark: " + psNode.label());
				}
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}
		
	}
	
	
	/**
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateAnticausative(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate anticausative...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^V/ >+(/^V/) (/^VP/=vp [> /^S/ $ @NP | >: /^SNS/ | >: /\\*GAP.\\*/ ] !<+(/^V/) (V $ /ARG./)) <: (__ !< __)");
			TregexMatcher matcher = tregex.matcher(psTree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString();
				boolean isAnticausative = lexEntry.contains("ANTICAUSATIVE");
				Tree vpNode = matcher.getNode("vp");
//				System.out.println("\Check: " + node);
//				System.out.println("Lexical entry is: " + lexEntry);
				if (isAnticausative && !vpNode.label().toString().matches(".*\\*(?:CP|AC)\\*.*")) {
					vpNode.label().setFromString(vpNode.label().toString() + "*AC*");
//					System.out.println("Mark: " + vpNode);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	
	/**
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateControl(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate control verbs...");

		try {
			
			// Subject control
			TregexPattern subjectTregex = TregexPattern.compile("V|V' >+(V|V') (V|V' > (VP=vp < /^V['P]?-C/ [> /^S/ $ NP-SJ-ARG1 | >: /^SNS.*\\*ARG1\\*/ ])) <: (__ !< __)");
			TregexMatcher subjectMatcher = subjectTregex.matcher(psTree);
			while (subjectMatcher.find()) {
				Tree node = subjectMatcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString().toLowerCase();
				String lexType = entry2type.containsKey(lexEntry) ? entry2type.get(lexEntry) : null;
				boolean isControl = lexType != null && lexType.matches(".*subj(?:ect)?_control.*");
				Tree vpNode = subjectMatcher.getNode("vp");
//				System.out.println("\nCheck: " + node);
//				System.out.println("Lexicon: " + lexEntry + " -> " + lexType);
				if (isControl && !vpNode.label().toString().matches(".*\\*(?:CONTROL|CP)\\*.*")) {
					vpNode.label().setFromString(vpNode.label().toString() + "*CONTROL*");
//					System.out.println("Mark: " + vpNode);
				}
			}
			
			// Object control
			TregexPattern objectTregexp = TregexPattern.compile("V|V' >+(V|V') (V|V'=v $ NP-DO-ARG2) <: (__ !< __)");
			TregexMatcher objectMatcher = objectTregexp.matcher(psTree);
			while (objectMatcher.find()) {
				Tree node = objectMatcher.getMatch();
				String lexEntry = derivTree.getNodeNumber(node.nodeNumber(psTree)).label().toString().toLowerCase();
				String lexType = entry2type.containsKey(lexEntry) ? entry2type.get(lexEntry) : null;
				boolean isControl = lexType != null && lexType.matches(".*obj_control.*");
				Tree vNode = objectMatcher.getNode("v");
//				System.out.println("\nCheck: " + node);
//				System.out.println("Lexicon: " + lexEntry + " -> " + lexType);
				if (isControl && !vNode.label().toString().matches(".*\\*(?:CONTROL|CP)\\*.*")) {
					vNode.label().setFromString(vNode.label().toString() + "*CONTROL*");
//					System.out.println("Mark: " + vNode);
				}
			}

		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Annotate pre-terminal nodes with their lexical type.
	 * 
	 * @param psTree
	 * @param derivTree
	 */
	private void annotateWithLexicalTypes(Tree psTree, Tree derivTree) {
//		System.out.println("\nAnnotate with lexical types...");
		List<Tree> leaves = derivTree.getLeaves();
		for (Tree leaf : leaves) {
			Tree parent = leaf.parent(derivTree);
			String lexEntry = parent.label().toString().toLowerCase();
			String lexType = null;
			if (entry2type.containsKey(lexEntry)) {
				lexType = entry2type.get(lexEntry);
			} else {
				if (lexEntry.endsWith("_external-name")) {
					lexType = "generic-noun-proper-form";
				} else if (lexEntry.endsWith("_external-card")) {
					lexType = "generic-cardinal-form";
				} else if (lexEntry.endsWith("_external-ord")) {
					lexType = "generic-ordinal-form";
				} else {
					lexType = null;
				}
			}
			Tree preterminal = psTree.getNodeNumber(parent.nodeNumber(derivTree));
//			System.out.println("Tag: " + preterminal + " <- " + lexType + " (" + lexEntry + ")");
			preterminal.label().setFromString(preterminal.label().toString() + "*lextype:" + lexType + "*");
		}
	}
	
	/*
	 * MWU handling 
	 */
	
	
	/**
	 * Flattens structure of complex CARDs.
	 * 
	 * @param tree
	 */
	private void flattenCARDs(Tree tree) {
//		System.out.println("\nFlatten complex CARD structures...");

		// Start by merging MWU CARDs, in case they haven't been already merged
		mergeMultiwordNodes(tree, "^CARD.*");
		
		try {
			
			// Preprocess (merge) MWU ADVs that modify the CARD, in case they haven't been already merged
			TregexPattern tregex = TregexPattern.compile("CARD <1 /^ADV/=l <2 CARD");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				mergeMultiwordNodes(matcher.getNode("l"), "^ADV.*");
			}

			// Flatten the structure
//			tregex = TregexPattern.compile("@CARD=t [ <1 CARD|ADV=l <2 CARD=r | <1 CARD=l [ <2 (CONJP <1 CONJ=c <2 CARD=r) | <2 (CARD <1 CONJ=c <2 CARD=r) ] ]");
			tregex = TregexPattern.compile("@CARD=t [ <1 CARD=l <2 CARD=r | <1 CARD=l [ <2 (CONJP <1 CONJ=c <2 CARD=r) | <2 (CARD <1 CONJ=c <2 CARD=r) ] ]");
			boolean treeAltered;
			do {
				treeAltered = false;
				matcher = tregex.matcher(tree);
				while (matcher.find()) {
//					System.out.println("\nMatch: " + matcher.getMatch());
					Tree tNode = matcher.getNode("t");
					Tree rNode = matcher.getNode("r");
					Tree lNode = matcher.getNode("l");
					Tree cNode = matcher.getNode("c");
					if (lNode.isPreTerminal() && rNode.isPreTerminal()) {
//						System.out.println("\nFlatten: " + matcher.getMatch());
						String newLeafLabel = "";
						newLeafLabel += lNode.firstChild().label().toString();
						if (cNode != null) {
							newLeafLabel += " " + cNode.firstChild().label().toString();
						}
						newLeafLabel += " " + rNode.firstChild().label().toString();
//						System.out.println("-> " + newLeafLabel);
						
//						LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
						List<Tree> newLeaf = new ArrayList<Tree>(1);
						newLeaf.add(treeFactory.newLeaf(newLeafLabel));
						tNode.setChildren(newLeaf);
						
						treeAltered = true;
					}
				}
			} while (treeAltered);
			
		} catch (ParseException e) {
			e.printStackTrace();
		}
		
	}

	/**
	 * Finds leaf nodes that are multiword and expands them into several nodes, one for each word in the multiword.
	 * <p>
	 * A leaf if considered to be a multiword if it contains a whitespace character.<br>
	 * The newly created nodes:
	 * <ul>
	 * <li>are placed as sister nodes beneath the node that was parent to the multiword
	 * <li>are labeled with the same label of the node that was parent to the multiword, with any markings removed
	 * <li>the parent of the added nodes is marked as being a multiword node
	 * </ul>
	 * TODO: Get the MWU mark from Properties
	 * 
	 * @param tree  the parse tree
	 */
	private void expandMultiwordLeaves(Tree tree) {
//		System.out.println("\nExpand multiword leaves...");
		List<Tree> leaves = tree.getLeaves();
//		System.out.println(leaves + " (" + leaves.size() + " leaves)");
		for (Tree leaf : leaves) {
			String label = leaf.label().toString();
			if (label.contains(" ")) {
				Tree leafParent = leaf.parent(tree);
				String mwuLabel = leafParent.label().toString().replaceFirst("\\*[^*]*\\*$", "");
				List<Tree> newChildren = new ArrayList<Tree>();
				for (String newWord : java.util.Arrays.asList(label.split(" "))) {
					List<Tree> newLeaf = new ArrayList<Tree>(1);
					newLeaf.add(treeFactory.newLeaf(newWord));
					newChildren.add(treeFactory.newTreeNode(mwuLabel, newLeaf));
				}
				leafParent.setChildren(newChildren);
				leafParent.label().setFromString(leafParent.label().toString() + "*MWU*");
			}
		}
	}
	

	/**
	 * Relabels PRS2 nodes in order to form MWU possessives.
	 * 
	 * @param tree
	 */
	private void fixPRS2(Tree tree) {
//		System.out.println("\nFix PRS2 nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("PRS2=r >2 (POSS-M=t <1 POSS=l)");
			TsurgeonPattern tsurgeon_step1 = Tsurgeon.parseOperation("relabel l POSS-M");
			TsurgeonPattern tsurgeon_step2 = Tsurgeon.parseOperation("relabel r POSS-M");
			TsurgeonPattern tsurgeon_step3 = Tsurgeon.parseOperation("relabel t POSS-M*MWU*");
			List<TsurgeonPattern> operations = new ArrayList<TsurgeonPattern>();
			operations.add(tsurgeon_step1);
			operations.add(tsurgeon_step2);
			operations.add(tsurgeon_step3);
			TsurgeonPattern tsurgeon = Tsurgeon.collectOperations(operations);
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
		} catch (ParseException e) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * 
	 * @param tree
	 */
	@SuppressWarnings("unused")
	private void mergeMultiwordNodes(Tree tree) {
		mergeMultiwordNodes(tree, ".*");
	}

	
	/**
	 * Merges the children of a multiword node into a single leaf.
	 * <p>
	 * Requires that multiword nodes be marked with a special tag (done previously by the <code>expandMultiwordLeaves</code> method).
	 * <p>
	 * TODO: How to merge nodes that have annotation (POS, morphology, lemma and/or IOB)?
	 * TODO: Get the MWU mark from Properties
	 *  
	 * @param tree
	 * @param categoryFilter  regexp pattern matching the categories to be merged (the regexp must match the whole string)
	 */
	private void mergeMultiwordNodes(Tree tree, String categoryFilter) {
//		System.out.println("\nMerge multiword nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*MWU\\*/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree multiwordNode = matcher.getMatch();
				boolean filter = multiwordNode.label().toString().matches(categoryFilter);
				if (filter && multiwordNode.numChildren() > 1) {
					List<Tree> children = multiwordNode.getChildrenAsList();
//					System.out.println("\nMerge:\n" + children + " (" + children.size() + ")");
					StringBuilder accumulator = new StringBuilder();
					for (Tree child : children) {
						accumulator.append(" " + child.getChild(0).label().toString());
					}
					String newString = accumulator.substring(1).trim();
//					System.out.println("\nNew leaf:\n" + newString);
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					List<Tree> newLeaf = new ArrayList<Tree>(1);
					newLeaf.add(treeFactory.newLeaf(newString));
					multiwordNode.setChildren(newLeaf);
					multiwordNode.label().setFromString(multiwordNode.label().toString().replaceAll("\\*MWU\\*", ""));
				}
//				System.out.println("\nNew tree:\n" + tree);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Replaces the whitespace in MWU by some other symbol.
	 * <p>
	 * This is only necessary if MWU nodes have been merged, as a way of keeping tokens and leaves aligned.
	 * <p>
	 * TODO: Get the symbol used to connect the MWU parts from Properties
	 * 
	 * @param tree
	 */
	private void fixSpacesInMultiwordNodes(Tree tree) {
		List<Tree> leaves = tree.getLeaves();
		for (Tree leaf : leaves) {
			if (leaf.label().toString().contains(" ")) {
				leaf.label().setFromString(leaf.label().toString().replaceAll(" +", "+"));
			}
		}
	}
		
	
	/**
	 * Restores the original orthography to the parse tree.
	 * <p>
	 * The exported Redwoods parse has all words in lowercase.
	 * This method restores the original orthography by replacing them with the corresponding token from the annotated sentence.
	 * <p>
	 * This method expects that the number of leaf nodes in the parse tree be equal to the number of tokens in the annotated sentence.
	 * Consequently, its requires that <code>expandMultiwordLeaves</code> and <code>detachPunctuation</code> be previously run over the parse tree.
	 * <p>
	 * This method is only concerned with restoring the original orthography.
	 * As such, it removes all annotation from the token, such as POS tags, morphologic features, lemma, etc.
	 * See <code>addPOSNodes</code> for a method that adds annotation from the sentence to the parse tree.
	 * <p>
	 * On the other hand, annotation that is created during the tokenization process, namely the markings for contraction expansion and vocalic alteration, are kept as part of the token.
	 * <p>
	 * TODO: Should tokenization annotation be kept? Should it be externally configurable?
	 * TODO: Assume that input has been pre-processed and remove the temporary fix for parenthesis in the annotated sentence.
	 * 
	 * @param tree       the parse tree
	 * @param annotated  the annotated sentence (a String in CINTIL format)
	 */
	private void restoreOrthography(Tree tree, String annotated) {
//		System.out.println("\nRestore orthography...");
		List<Tree> leaves = tree.getLeaves();
		String[] tokens = annotated.split(" +");
//		System.out.println(leaves + " (" + leaves.size() + " leaves)");
//		System.out.println(java.util.Arrays.asList(tokens) + " (" + tokens.length + " tokens)");
		int i = 0;
		for (Tree leaf : leaves) {
			String token = tokens[i];
			if (token.contains("(") || token.contains(")")) {
				// FIXME: Temporary fix for skipping parenthesis in input
			} else {
				token = token.replaceFirst("\\*//", "/");     // Remove right-adjoin spacing mark
				token = token.replaceFirst("^\\\\\\*", "");   // Remove left-adjoin spacing mark
				token = token.replaceFirst("(.+?)/.+", "$1"); // Remove annotation
//				token = token.replaceFirst("_$", "");         // Remove contraction expansion mark
//				token = token.replaceFirst("_#", "");         // Remove vocalic alteration mark
				leaf.label().setFromString(token);
			}
			i++;
		}
	}
	
	
	/**
	 * Replaces IOB multiword units by their type.
	 * <p>
	 * At this stage, leafs and tokens are no longer aligned.
	 * Requires <code>addAnnotationFromSentence</code>.
	 * 
	 * @param tree       the parse tree
	 * @param annotated  the annotated sentence (a String in CINTIL format)
	 */
	@Deprecated
	private void replaceIOBByTypes(Tree tree, String annotated) {
//		System.out.println("\nReplace IOB MWU by their types...");
		
		// Use the IOB information to detect the beginning of multiwords
		Pattern iobTypePattern = Pattern.compile(".+\\[B-(.+)]");
		List<Tree> leaves = tree.getLeaves();
//		System.out.println("Leaves: " + leaves);
		String[] tokens = annotated.split(" +");
//		System.out.println("CINTIL: " + annotated);
		int i = 0;
		for (Tree leaf : leaves) {
			String leafLabel = leaf.label().toString();
			String token = tokens[i];
			Matcher m = iobTypePattern.matcher(token);
//			System.out.println(leafLabel + " : " + token);
			if (m.matches()) {
//				System.out.println("> Found B: " + m.group());
				leaf.label().setFromString("*" + m.group(1) + "*");
			}
			i += leafLabel.split(" +").length;
		}
		
		// TIMEX
		// N <: /(?i:janeiro|fevereiro|março|abril|maio|junho|julho|agosto|setembro|outubro|novembro|dezembro)/
		// N <: /(?i:(?:segunda|terça|quarta|quinta|sexta)(?:-feira)?|sábado|domingo)/
	}

	/**
	 * Adds a lexical type tag to the leaves.
	 * <p>
	 * The tag is taken from the preterminal node, where it was placed by <code>annotateWithLexicalTypes</code>.
	 *  
	 * @param tree
	 */
	private void addLexicalTypes(Tree tree) {
//		System.out.println("\nAdd lexical types...");
		try {
			TregexPattern lextypeMarkTregex = TregexPattern.compile("/\\*lextype:/ << (__=leaf !< __ !> DETACHPNT)");
			TregexMatcher lextypeMarkMatcher = lextypeMarkTregex.matcher(tree);
			List<Tree> markedNodes = new ArrayList<Tree>();
			while (lextypeMarkMatcher.find()) {
				Tree node = lextypeMarkMatcher.getMatch();
				markedNodes.add(node);
				Tree leaf = lextypeMarkMatcher.getNode("leaf");
				String lextype = node.label().toString().replaceFirst(".*\\*lextype:([^*]+)\\*.*", "$1");
//				System.out.println("Leaf: " + leaf + " <- " + lextype);
				leaf.label().setFromString(leaf.label().toString() + ":" + lextype);
			}
			// Cleanup markings
			for (Tree node : markedNodes) {
				node.label().setFromString(node.label().toString().replaceFirst("\\*lextype:[^*]+\\*", ""));
			}
			// Add lexical type for punctuation nodes
			TregexPattern pntTregex = TregexPattern.compile("DETACHPNT");
			TregexMatcher pntMatcher = pntTregex.matcher(tree);
			while (pntMatcher.findNextMatchingNode()) {
				Tree leaf = pntMatcher.getMatch().firstChild();
				leaf.label().setFromString(leaf.label().toString() + ":pnt");
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Adds feature bundles to the leaves.
	 * <p>
	 * The feature bundles are taken from the annotated sentence.
	 * Therefore, this method expects that the number of leaf nodes in the parse tree be equal to the number of tokens in the annotated sentence.
	 * Consequently, its requires that <code>expandMultiwordLeaves</code> and <code>detachPunctuation</code> be previously run over the parse tree.
	 * <p>
	 * In addition, this method expects the annotation of the sentence to follow the CINTIL/LX-Suite format.
	 * Each token must be in one of these formats:<br>
	 * <ul>
	 * <li><code>token/LEMMA/POS#features[IOB]</code>
	 * <li><code>token/POS#features[IOB]</code>
	 * <li><code>token/POS[IOB]</code>
	 * </ul>
	 * <p>
	 * TODO: Decide which information should be added (POS, morphologic features, lemma, IOB, etc) and how to make it externally configurable
	 * 
	 * @param tree       the parse tree
	 * @param annotated  the annotated sentence (a String in CINTIL format)
	 */
	private void addFeatureBundles(Tree tree, String annotated) {
//		System.out.println("\nAdd feature bundles...");
		List<Tree> leaves = tree.getLeaves();
		String[] tokens = annotated.split(" +");
//		System.out.println(leaves + " (" + leaves.size() + " leaves)");
//		System.out.println(java.util.Arrays.asList(tokens) + " (" + tokens.length + " tokens)");
		
		Pattern iobPattern = Pattern.compile("^(.+)\\[(.+)\\]$");
		Pattern featuresPattern = Pattern.compile("^(.+)#([^/].+)$");
		Pattern posPattern = Pattern.compile("^(.+)/(.+)$");
		Pattern lemmaPattern = Pattern.compile("^(.+)/(.+)$");

		int i = 0;
		for (Tree leaf : leaves) {
			String token = tokens[i];
			
			// Extract IOB
			String iob = null;
			Matcher m = iobPattern.matcher(token);
			if (m.matches()) {
				token = m.group(1);
				iob = m.group(2);
			}

			// Extract inflection features
			String features = null;
			m = featuresPattern.matcher(token);
			if (m.matches()) {
				token = m.group(1);
				features = m.group(2);
			}

			// Extract POS
			String pos = null;
			m = posPattern.matcher(token);
			if (m.matches()) {
				token = m.group(1);
				pos = m.group(2);
			}

			// Extract lemma
			String lemma = null;
			m = lemmaPattern.matcher(token);
			if (m.matches()) {
				token = m.group(1);
				lemma = m.group(2);
				lemma = lemma.toUpperCase(); // ensure uppercase
			}

			// Creates the feature bundle
			//System.out.println(token + " + " + lemma + " + " + pos + " + " + features + " + " + iob);
			String bundle = null;
			if (pos != null) {
				if (bundle != null) { bundle += pos; } else { bundle = pos;	}
			}
			if (lemma != null) {
				if (bundle != null) { bundle += ":" + lemma; } else { bundle = lemma; }
			}
			if (features != null) {
				if (bundle != null) { bundle += ":" + features; } else { bundle = features;	}
			}
			if (iob != null) {
				if (bundle != null) { bundle += ":" + iob; } else { bundle = iob; }
			}

			// Add bundle to leaf
			if (bundle != null) {				
				leaf.label().setFromString(leaf.label().toString() + "/" + bundle);
			}
			
			i++;
		}
	}
	
	
	/**
	 * Flattens unary chains of the same constituent.
	 * <p>
	 * The Redwoods export of the phrase structure tree has unary chains of the same constituent which typically correspond to the application of different lexical rules in the HPSG grammar.
	 * These unary chains are not meaningful for the constituent representation and are therefore collapsed into a single constituent.
	 * <p>
	 * For instance:<br>
	 * (NP (N (N (N (gato))))) -> (NP (N gato))
	 * <p>
	 * Only the constituent tag (base category) is taken into account when determining if two nodes match.
	 * That is, syntactic function and semantic role tags are ignored.
	 * <p>
	 * For instance:<br>
	 * (D-SP (D-SP (D (D a)))) -> (D-SP a)
	 * <p>
	 * Notes:
	 * <ul>
	 * <li>The topmost node of the chain is the one that is kept in the final result.
	 * <li>MWU markings are preserved (moved upwards from the excised node).
	 * </ul>
	 * <p>
	 * TODO: Get the MWU mark from Properties
	 * 
	 * @param tree  the parse tree
	 */	
	private void flattenUnaryChains(Tree tree) {
//		System.out.println("\nFlatten unary chains...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^([^-]+)/#1%c=n <: (/^([^-]+)/#1%c == __=x < __)");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("n");
				Tree exciseNode = matcher.getNode("x");
				boolean mwuExcised = exciseNode.label().toString().contains("*MWU*");
				boolean gaplExcised = exciseNode.label().toString().contains("*GAPl*");
				boolean gaprExcised = exciseNode.label().toString().contains("*GAPr*");
				topNode.setChildren(exciseNode.getChildrenAsList());
				if (mwuExcised) {
					topNode.label().setFromString(topNode.label().toString() + "*MWU*");
				}
				if (gaplExcised) {
					topNode.label().setFromString(topNode.label().toString() + "*GAPl*");
				}
				if (gaprExcised) {
					topNode.label().setFromString(topNode.label().toString() + "*GAPr*");
				}
				matcher = tregex.matcher(tree);
			}

		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/*
	 * Fixes/Hacks
	 */


	/**
	 * Relabels CP nodes that immediately dominate a single S or SNS, giving them the label of that child.   
	 * <p>
	 * This relabeling is meant to be followed by <code>flattenUnaryChains</code>, which will then remove the relabeled node.
	 * 
	 * @param tree  the parse tree
	 */
	private void relabelUnaryCP(Tree tree) {
//		System.out.println("\nRemove extra CP...");
		try {
			TregexPattern tregex = TregexPattern.compile("CP=t <: /^(S|SNS)(\\*.*)?$/=c");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				Tree childNode = matcher.getNode("c");
				topNode.label().setFromString(childNode.label().toString());
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Some methods alter the node labels. If the changed node is co-indexed with another ndoe, the label of that node must be changed as well. 
	 * 
	 * @param tree  the parse tree
	 * @param from  the old label
	 * @param to    the new label
	 */
	private void applyChangeToCoindexedNodes(Tree tree, String from, String to) {
//		System.out.println("\nApplying change to GAPs...");
		try {
			TregexPattern tregex = TregexPattern.compile(from + "=n <: /\\*GAP\\*/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				node.label().setFromString(to);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Adds a X-bar mark to POS tags that occur in phrasal nodes.
	 * <p>
	 * TODO: Are these all the possible POS tags? Externalize the list.
	 * 
	 * @param tree
	 */
	private void addXbarTag(Tree tree) {
//		System.out.println("\nAdd X-bar tag to POS nodes in phrasal position...");
		try {
			TregexPattern tregex = TregexPattern.compile("@A|ADV|ART|C|CARD|CL|CONJ|D|DEM|ITJ|N|ORD|P|PERCENT|PNT|POSS|PPA|PRS|QNT|REL|V=x");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getNode("x");
				if (node.isPhrasal()) {
					String oldLabel = node.label().toString();
					String newLabel = oldLabel.replaceFirst("(\\w+)", "$1'");
//					System.out.println("Relabel: " + oldLabel + " -> " + newLabel);
					if (newLabel.matches(".+_\\d+$")) {
						applyChangeToCoindexedNodes(tree, oldLabel, newLabel);
					}
					node.label().setFromString(newLabel);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/*
	 * Methods for handling punctuation
	 */
	
	
	/**
	 * Detaches punctuation symbols from the words to which they are attached and creates a new node to hold the detached punctuation symbol.
	 * <p>
	 * The detached symbol are simply placed in a new node created as a sister to the word.
	 * This is done to:
	 * <ul>
	 * <li>easily identify where punctuation symbols occur in the tree
	 * <li>ensure that the number of leaves in the tree and to the number of tokens in the sentence are the same
	 * </ul>
	 * Note, however, that this naive placement of the nodes is only temporary.
	 * The detached punctuation nodes will be relabeled and moved to suitable positions in the tree by other methods.
	 * <p>
	 * This method requires access to the annotated sentence so that, when it finds a word followed by a period, it can decide if it should consider that as being an abbreviation or it should detach the period.
	 * <p>
	 * TODO: The temporary label for detached punctuation could bear special symbols to ensure its uniqueness. Externalize this label so that it can be seen by the other methods.
	 * TODO: Try to factor out the common code for handling left-adjoining and right-adjoining punctuation
	 *  
	 * @param tree       the parse tree
	 * @param annotated  the annotated sentence (a String in CINTIL format)
	 */
	private void detachPunctuation(Tree tree, String annotated) {
//		System.out.println("\nDetach punctuation...");
		
		String nodeLabel = "DETACHPNT";
		
		List<Tree> leaves = tree.getLeaves();
		String[] tokens = annotated.split(" +");

//		System.out.println(leaves + " (" + leaves.size() + " leaves)");
//		System.out.println(java.util.Arrays.asList(tokens) + " (" + tokens.length + " tokens)");
		
		Pattern detachLeftRegexp = Pattern.compile("^(['\"])(.+)$");
		Pattern detachRightRegexp = Pattern.compile("^(.+?)(\\.\\.\\.|[,.!?;:'\"])$");
		
		/*
		 * Scans through leaves (i) and their corresponding tokens (j). 
		 * For a leaf i to be a valid target for left detachment, the j token must be a punctuation token. 
		 * For a leaf i to be a valid target for right detachment, the j+1 token must be a punctuation token. 
		 */
		int i = 0, j = 0;
		while (i < leaves.size()) {
			Tree leaf = leaves.get(i);
			Label label = leaf.label();

//			System.out.println("Check: " + label);

			// Check for left-adjoining punctuation for detachment
			Matcher matcherLeft = detachLeftRegexp.matcher(label.toString());
			while (matcherLeft.matches() && tokens[j].contains("/PNT[")) {
//				System.out.println("Detach left: " + matcherLeft.group(1) + "|" + matcherLeft.group(2));
				
				// Relabel word
				label.setFromString(matcherLeft.group(2));
//				System.out.println("Relabel node to: " + label);
				
				// Create punctuation node
				Tree newPunctNode = string2tree("(" + nodeLabel + " " + matcherLeft.group(1) + ")");
//				System.out.println("Create node: " + newPunctNode);
				
				// Insert the newly created punctuation node as a left sister of the word node
				Tree leafParent = leaf.parent(tree);
				int leafIndex = leafParent.indexOf(leaf);
				leafParent.addChild(leafIndex, newPunctNode);
//				System.out.println("Resulting node: " + leafParent);
				
				// Increment token index (to maintain leaf-token alignment)
				j++;
				
				//
				matcherLeft = detachLeftRegexp.matcher(label.toString());
			}

			// Check for right-adjoining punctuation for detachment
			Matcher matcherRight = detachRightRegexp.matcher(label.toString());
			while (matcherRight.matches() && j+1 < tokens.length && tokens[j+1].contains("/PNT[")) {
//				System.out.println("Detach right: " + matcherRight.group(1) + "|" + matcherRight.group(2));
				
				// Relabel word
				label.setFromString(matcherRight.group(1));
//				System.out.println("Relabel node to: " + label);
					
				// Create punctuation node
				Tree newPunctNode = string2tree("(" + nodeLabel + " " + matcherRight.group(2) + ")");
//				System.out.println("Create node: " + newPunctNode);
				
				// Insert the newly created punctuation node as a right sister of the word node
				Tree leafParent = leaf.parent(tree);
				int leafIndex = leafParent.indexOf(leaf);
				leafParent.addChild(leafIndex+1, newPunctNode);
//				System.out.println("Resulting node: " + leafParent);
				
				// Increment token index (to maintain leaf-token alignment)
				j++;
				
				//
				matcherRight = detachRightRegexp.matcher(label.toString());					
			}
			
			// Increment leaf and token indexes
			i++; 
			j++;
			
		}
		
	}
	
	
	/**
	 * Raises punctuation symbols as far up the tree as possible, without entering into new constituents.
	 * <p>
	 * This method makes use of three auxiliary methods which are repeatedly applied as long as the tree changes:
	 * <ul>
	 * <li><code>raiseMatchingPunctuation</code>
	 * <li><code>raisePunctuationOnLeft</code>
	 * <li><code>raisePunctuationOnRight</code>
	 * </ul>
	 * <p>
	 * TODO: Some very particular cases may be handled wrongly. In particular, if matching punctuation and non-matching punctuation occur in the same constituent.
	 *  
	 * @param tree
	 */
	private void raisePunctuation(Tree tree) {
//		System.out.println("\nRaise punctuation...");
		boolean treeChanged = false;
		do {
			treeChanged = raiseMatchingPunctuation(tree);
			if (treeChanged) {
				continue;
			}
			treeChanged = raisePunctuationOnRight(tree);
			if (treeChanged) {
				continue;
			}
			treeChanged = raisePunctuationOnLeft(tree);
			if (treeChanged) {
				continue;
			}
		} while (treeChanged);
	}
	
	
	/**
	 * Remove extraneous syntactic function and semantic role tags.
	 * 
	 * @param tree
	 */
	private void fix_removeExtraSynSemTags(Tree tree) {
		try {
			TregexPattern tregex = null;
			TregexMatcher matcher = null;
			
			// Coordinations
			// Only two cases, both with just two coordinated constituents.
			// Won't bother for now to make it work for more that two coordinated constituents.
			tregex = TregexPattern.compile("/^[^\\/]+-/=x <1 ~x <2 (~x <1 CONJ|PNT <2 ~x)");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				String newLabel = node.label().toString().replaceFirst("-[^/]+", "");
				node.getChild(0).label().setFromString(newLabel);
				node.getChild(1).label().setFromString(newLabel);
				node.getChild(1).getChild(1).label().setFromString(newLabel);
			}
			
			// Sisters of PNT
			tregex = TregexPattern.compile("/^[^\\/]+-/=x $ PNT > /^[^\\/]+-/");
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				node.label().setFromString(node.label().toString().replaceFirst("-[^/]+", ""));
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_Modifier2Specifier(Tree tree) {
//		System.out.println("\nFix POSS-M and QNT-M before nouns...");
		try {
			TregexPattern tregex = TregexPattern.compile("/(?:QNT|POSS)-M/ $++ N|N'");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				node.label().setFromString(node.label().toString().replaceFirst("-M", "-SP"));
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * Inserts preterminal (part-of-speech) tags between phrasal nodes and leafs.
	 * <p>
	 * TODO: Get the leaf name for null nodes from Properties.
	 * 
	 * @param tree
	 */
	private void fix_addLexicalPreterminalTags(Tree tree) {
//		System.out.println("\nFIX: Add pre-terminal (POS) tags...");
		try {
			TregexPattern tregex = TregexPattern.compile("@NP|ADVP=t <: (__=w !< __)");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				Tree leafNode = matcher.getNode("w");
				if (topNode.isPreTerminal() && !leafNode.label().toString().matches("\\*(NULL|ELLIPSIS|GAP|TOUGH)\\*")) {
					String pos = "XXX";
					// Choose the POS tag
					if (topNode.label().toString().startsWith("NP")) {
						String word = leafNode.label().toString().toLowerCase().replaceAll("/.+", "");
						if (word.matches("quem?")) {
							pos = "REL";
						} else if (word.matches("[et]u|el[ea]s?|[nv]ós|mim|[st]i|você")) {
							pos = "PRS";
						} else if (word.matches("-?(?:[smt]e|[nv]os?|lhes?|(?:n|l|m|t|lh)?[oa]s?)")) {
							pos = "CL";
						}
					} else {
						pos = "ADV";
					}
					// Add the new node
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					Tree[] newChildren = {treeFactory.newTreeNode(pos, topNode.getChildrenAsList())};
					topNode.setChildren(newChildren);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * Fix for v3:
	 * Coordinated CP nodes have syn/sem tags on their labels (only the top node should).
	 * 
	 * @param tree
	 */
	private void fix_removeFunctionFromCoordinatedCPs(Tree tree) {
//		System.out.println("\nFIX: Remove function from coordinated CPs...");
		try {
			TregexPattern tregex = TregexPattern.compile("/CP-M-/ !> /CP-M-/ <+(/CP-M-/) /CP-M-/=x");
			TregexMatcher matcher = tregex.matcher(tree);
			List<Tree> nodes = new ArrayList<Tree>();
			while (matcher.find()) {
//				System.out.println("Fix: " + matcher.getMatch());
				nodes.add(matcher.getNode("x"));
			}
			for (Tree node : nodes) {
				node.label().setFromString("CP");
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Fix for v3:
	 * 
	 * A: 
	 * C: que, de o que, como, quanto
	 * 
	 * @param tree
	 */
	private void fix_renameComparatives(Tree tree) {
//		System.out.println("\nFIX: Rename comparatives...");
		try {
			TregexPattern tregex = TregexPattern.compile("CP-C-ARG2 <1 C|C' $-- A|A'|ADV|ADV'");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getMatch();
				topNode.label().setFromString("CONJP-C-ARG2");
				Tree headNode = topNode.firstChild();
				headNode.label().setFromString(headNode.label().toString().replaceFirst("C", "CONJ"));
				if (headNode.isPrePreTerminal()) {
					for (Tree child : headNode.getChildrenAsList()) {
						child.label().setFromString("CONJ");
					}
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_removeXbarOverX(Tree tree) {
		try {
			TregexPattern tregex = TregexPattern.compile("/^(.+)'/#1%c=x <: /^(.+)/#1%c");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("excise x x");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void fix_sns2s(Tree tree) {
		try {
			TregexPattern tregex = TregexPattern.compile("SNS=t <2 (PNT <: /^[.?!:']+/)");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("relabel t FIXSNS");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);

			tregex = TregexPattern.compile("SNS=t < FIXSNS");
			tsurgeon = Tsurgeon.parseOperation("relabel t FIXSNS");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);

			tregex = TregexPattern.compile("FIXSNS=t");
			tsurgeon = Tsurgeon.parseOperation("relabel t S");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);

		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * Fix several problems in v3 where nodes bear the wrong label.
	 * 
	 * NB: Leafs may bear feature bundles, hence the optional "/" following the word.
	 * 
	 * @param tree
	 */
	private void fix_renameNodes(Tree tree) {
		try {
			// a, o: D-SP -> ART-SP
			TregexPattern tregex = TregexPattern.compile("D-SP=t <: /^(?i:[ao]s?)\\/?/");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("relabel t ART-SP");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
			
			// algum, nenhum, outro, qualquer : D -> QNT
			tregex = TregexPattern.compile("@D=t <: /^(?i:(?:algu|nenhu)(?:m|ns|mas?)|outr[oa]s?|qua(?:l|is)quer)\\/?/");
			tsurgeon = Tsurgeon.parseOperation("relabel t QNT");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
			
			// muito, imenso, vários : A -> QNT
			tregex = TregexPattern.compile("@A=t <: /^(?i:(?:muit|imens)[oa]s?|vári[oa]s)\\/?/");
			tsurgeon = Tsurgeon.parseOperation("relabel t QNT");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);

			// primeiro, segundo, terceiro, quarto, etc. : A -> ORD
			tregex = TregexPattern.compile("@A=t <: /^(?i:(?:primeir|segund|terceir|quart|quint|sext|sétim|oitav|non|décim)[oa]|(?:vig|tri)ésim[oa])\\/?/");
			tsurgeon = Tsurgeon.parseOperation("relabel t ORD");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);

			// V+SJ -> V'
			// NB: The TSurgeon relabel operation does not accept the quote symbol
			tregex = TregexPattern.compile("/V\\+SJ/=t");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.findNextMatchingNode()) {
				Tree node = matcher.getMatch();
				node.label().setFromString(node.label().toString().replaceFirst("V\\+SJ", "V'"));
			}

		} catch (ParseException e) {
			e.printStackTrace();
		}		
	}
	
	
	/**
	 * Fix for a problem in v3.
	 * 
	 * @param tree
	 */
	private void remove_D_over_CARDorA(Tree tree) {
//		System.out.println("\nRemove D over a single CARD or A...");
		try {
			TregexPattern tregex = TregexPattern.compile("@D|D'=n <: (@CARD|CARD'|A|A' < __)");
			TsurgeonPattern tsurgeon = Tsurgeon.parseOperation("excise n n");
			Tsurgeon.processPattern(tregex, tsurgeon, tree);
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Fix for a problem in v3.
	 * 
	 * @param tree
	 */
	private void addSynFuncToCARDs(Tree tree) {
//		System.out.println("\nAdd syntactic functions to some CARD structures...");
		try {
			TregexPattern tregex = TregexPattern.compile("CARD' < (ADV|ADV'=a $+ CARD|CARD')");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree advNode = matcher.getNode("a");
				advNode.label().setFromString(advNode.label().toString() + "-M");
			}
			
			tregex = TregexPattern.compile("@NP|N' < (CARD|CARD'=c $+ N|N')");
//			tregex = TregexPattern.compile("@NP|N' [ < (CARD|CARD'=c $+ N|N') | <: CARD|CARD'=c ]"); // Add alternative case for when ELLIPSIS aren't added?
			matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree cardNode = matcher.getNode("c");
				cardNode.label().setFromString(cardNode.label().toString() + "-SP");
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Fix...
	 * 
	 * @param tree
	 */
	private void addSynFuncToSComplement(Tree tree) {
//		System.out.println("\nAdd syntactic function to a S complement structure...");
		try {
			TregexPattern tregex = TregexPattern.compile("__ < (CONJ|CONJ' $+ S|SNS=s)");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree sNode = matcher.getNode("s");
				sNode.label().setFromString(sNode.label().toString() + "-C");
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * 
	 * @param tree
	 */
	private void preprocessParentheticalModifierPunctuation(Tree tree) {
//		System.out.println("\nPreprocess punctuation in parenthetical modifiers...");
		try {
			TregexPattern tregex = null;
//			tregex = TregexPattern.compile("__ <1 (__ <<- (DETACHPNT=p <: /,/)) <2 __=r"); // search high (preference for last modifier in ambivalent cases)
			tregex = TregexPattern.compile("DETACHPNT=p <: /,/ >>- (__  $++ __=r)"); // search low (preference for first modifier in ambivalent cases)
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree punctNode = matcher.getNode("p");
				Tree searchNode = matcher.getNode("r");
//				System.out.println("\nMatch: " + matcher.getMatch());
//				System.out.println("Start: " + searchNode);
				
				// Starting from the second child (R), search down the left side for a modifier
				boolean finished = false;
				while (!finished) {
					finished = searchNode.label().toString().contains("*M*");
					if (!finished) {
						searchNode = searchNode.firstChild();
						finished = (searchNode == null);
					}
				}
				
				// Found modifier. Check for matching punctuation.
				if (searchNode != null) {
					Tree modifierNode = searchNode;
//					System.out.println("Found modifier: " + modifierNode);
					
					// Search down the right side for matching detachable punctuation
					finished = false;
					while (!finished) {
						finished = searchNode.label().toString().matches("DETACHPNT");
						if (!finished) {
							searchNode = searchNode.lastChild();
							finished = (searchNode == null);
						}
					}
					
					// Move the punctuation.
					if (searchNode != null) {
						
						// Remove punctuation from where it is now
						Tree punctParent = punctNode.parent(tree);
						punctParent.removeChild(punctParent.numChildren() - 1);

						// Remove the modifier mark from the node that will hold the punctuation
						modifierNode.label().setFromString(modifierNode.label().toString().replaceFirst("\\*M\\*", ""));

						// Decide if we need to create a new node to hold the punctuation node
						boolean nodeHasSeveralChildren = modifierNode.numChildren() > 1;
						boolean nodeIsPreTerminal = modifierNode.isPreTerminal();
						if (nodeHasSeveralChildren || nodeIsPreTerminal) {
//							LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
							Tree newNode = treeFactory.newTreeNode(modifierNode.label().toString(), modifierNode.getChildrenAsList());
							List<Tree> newChildren = new ArrayList<Tree>(2);
							newChildren.add(punctNode);
							newChildren.add(newNode);
							modifierNode.setChildren(newChildren);
						} else {
							modifierNode.addChild(0, punctNode);
						}
						
						// Raise the matching punctuation now
						raiseMatchingPunctuation(modifierNode);
						
					}

				}
				
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Helper method that determines if a new node should be created to hold a raised punctuation node.
	 * 
	 * @param tree
	 * @return
	 */
	private boolean raisingRequiresNewNode(Tree tree) {
		boolean result = false;
		try {
			boolean nodeIsMarked = tree.label().toString().contains("*");
			boolean nodeHasSeveralChildren = tree.numChildren() > 1;
			boolean nodeIsPreTerminal = tree.isPreTerminal();
			boolean nodeIsNullVerb = TregexPattern.compile("@VP <: /^.+-PRD/ !<: /^VP?/").matcher(tree).matches();
//			System.out.println(
//					"nodeIsMarked: " + nodeIsMarked + "\n" + 
//					"nodeHasSeveralChildren: " + nodeHasSeveralChildren + "\n" + 
//					"nodeIsPreTerminal: " + nodeIsPreTerminal + "\n" + 
//					"nodeIsNullVerb: " + nodeIsNullVerb);
			result = (nodeIsMarked || nodeHasSeveralChildren || nodeIsPreTerminal || nodeIsNullVerb);
		} catch (ParseException e) {
			e.printStackTrace();
		}
		return result;
	}
	
	/**
	 * Raises paired punctuation symbols up the tree.
	 * <p>
	 * Paired punctuation is defined as two matching symbols that occur as the leftmost node of the first branch and rightmost node of the second branch of a tree.
	 * The method moves them as far up the tree as possible while maintaining their pairing.
	 * <p>
	 * TODO: Make it work for enclosing parenthesis (where the symbols are not equal)
	 * TODO: Get the punctuation label from Properties
	 * TODO: It assumes <code>flattenUnaryChains</code> has been run. I'd like to not need this.
	 * 
	 * @param tree
	 */
	private boolean raiseMatchingPunctuation(Tree tree) {
//		System.out.println("\nRaise paired punctuation...");
		
		boolean treeChanged = false;

		try {
			TregexPattern tregex = TregexPattern.compile("__ <<, (DETACHPNT=l <: __=p) <<- (DETACHPNT=r <: ~p)");
			TregexMatcher matcher = tregex.matcher(tree);
			Tree leftNode = null;
			Tree rightNode = null;
			
			// Find lowest match
			Tree lowest = null;
			while (matcher.find()) {
				Tree match = matcher.getMatch();
				if (lowest == null || lowest.dominates(match)) {
					lowest = match;
					leftNode = matcher.getNode("l");
					rightNode = matcher.getNode("r");
				}
			}
//			System.out.println("Lowest: " + lowest);
			
			// Raise lowest punctuation pair
			if (lowest != null) {
				
//				System.out.println("\nModify:\n" + tree);
//				System.out.println("\nT: " + topNode + "\nL: " + leftNode + "\nR: " + rightNode);
				
				// Remove punctuation nodes from their current location
				treeChanged = true;
				Tree leftParent = leftNode.parent(tree);
				leftParent.removeChild(0);
				Tree rightParent = rightNode.parent(tree);
				rightParent.removeChild(rightParent.numChildren() - 1);
				
				// Find the node where the punctuation will be attached
				Tree topNode = lowest.upperMostUnary(tree);
//				System.out.println("Attach: " + topNode);
				
				// Raise punctuation node
				if (raisingRequiresNewNode(topNode)) {
					// Create new structure
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					Tree newNode = treeFactory.newTreeNode(topNode.label().toString().replaceFirst("([^-/]+)(?:-[^-/*]+)*", "$1"), topNode.getChildrenAsList());
					leftNode.label().setFromString("PNT");
					rightNode.label().setFromString("PNT");
					topNode.setChildren(new Tree[]{leftNode, newNode, rightNode});
					// If the top node had ellipsis or null subject marks, we remove them
					topNode.label().setFromString(topNode.label().toString().replaceAll("\\*[^*]*\\*$", ""));
				} else {
					leftNode.label().setFromString("PNT");
					topNode.addChild(0, leftNode);
					rightNode.label().setFromString("PNT");
					topNode.addChild(rightNode);
				}
				
			}
			
		} catch (ParseException e) {
			e.printStackTrace();
		}
		return treeChanged;
	}
	
	
	/**
	 * Raises punctuation symbols on the left up the tree.
	 * <p>
	 * TODO: Get the punctuation label from Properties
	 * TODO: It assumes <code>flattenUnaryChains</code> has been run. I'd like to not need this.
	 * 
	 * @param tree
	 */
	private boolean raisePunctuationOnLeft(Tree tree) {
//		System.out.println("\nRaise a punctuation symbol on the left...");
		boolean treeChanged = false;
		try {
			TregexPattern tregex = TregexPattern.compile("__=t  <<, DETACHPNT=p");
			TregexMatcher matcher = tregex.matcher(tree);
			if (matcher.find()) {
//				System.out.println("\nModify:\n" + tree);
				treeChanged = true;
				Tree topNode = matcher.getNode("t");
				Tree punctNode = matcher.getNode("p");
//				System.out.println("\nT: " + topNode + "\nP: " + punctNode);
				
				// Remove punctuation node
				Tree punctParent = punctNode.parent(tree);
				punctParent.removeChild(punctParent.indexOf(punctNode));

				// Raise punctuation node
				if (raisingRequiresNewNode(topNode)) {
					// Create new structure
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					Tree newNode = treeFactory.newTreeNode(topNode.label().toString().replaceFirst("([^-/]+)(?:-[^-/*]+)*", "$1"), topNode.getChildrenAsList());
					punctNode.label().setFromString("PNT");
					topNode.setChildren(new Tree[]{punctNode, newNode});
					// If the top node had ellipsis or null subject marks, we remove them
					topNode.label().setFromString(topNode.label().toString().replaceAll("\\*[^*]*\\*$", ""));
				} else {
					punctNode.label().setFromString("PNT");
					topNode.addChild(0, punctNode);					
				}

			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
		return treeChanged;
	}
	
	
	/**
	 * Raises punctuation symbols on the right up the tree.
	 * <p>
	 * TODO: Get the punctuation label from Properties
	 * TODO: It assumes <code>flattenUnaryChains</code> has been run. I'd like to not need this.
	 * 
	 * @param tree
	 */
	private boolean raisePunctuationOnRight(Tree tree) {
//		System.out.println("\nRaise a punctuation symbol on the right...");
		boolean treeChanged = false;
		try {
			TregexPattern tregex = TregexPattern.compile("__=t  <<- DETACHPNT=p");
			TregexMatcher matcher = tregex.matcher(tree);
			if (matcher.find()) {
//				System.out.println("\nModify:\n" + tree);
				treeChanged = true;
				Tree topNode = matcher.getNode("t");
				Tree punctNode = matcher.getNode("p");
//				System.out.println("\nT: " + topNode + "\nP: " + punctNode);
				
				// Remove punctuation node
				Tree punctParent = punctNode.parent(tree);
				punctParent.removeChild(punctParent.indexOf(punctNode));

				// Raise punctuation node
				if (raisingRequiresNewNode(topNode)) {
				// Create new structure
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					Tree newNode = treeFactory.newTreeNode(topNode.label().toString().replaceFirst("([^-/]+)(?:-[^-/*]+)*", "$1"), topNode.getChildrenAsList());
					punctNode.label().setFromString("PNT");
					topNode.setChildren(new Tree[]{newNode, punctNode});
					// If the top node had ellipsis or null subject marks, we remove them
					topNode.label().setFromString(topNode.label().toString().replaceAll("\\*[^*]*\\*$", ""));
				} else {
					punctNode.label().setFromString("PNT");
					topNode.addChild(punctNode);
				}
				
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
		return treeChanged;
	}

	
	/**
	 * Moves punctuation symbols into their place in coordination structures that have an explicit conjunction.
	 * <p>
	 * This method requires that <code>detachPunctuation</code> be previously run to separate punctuation symbols from words.
	 * <p>
	 * This method expects the following structure:<br>
	 * (X (Y ... (DETACHPNT ,)) (X (Y ...) (X (CONJ|PNT ...) (Y ...))))
	 * <p>
	 * The punctuation node is moved into the same position that it would appear in if it were a conjunction.
	 * This requires adding an extra "X" node to hold the punctuation (on the left) and the remainder of the coordination (on the right).
	 * Apart from the structure, the method ensures that all Y nodes be equal. 
	 * In addition, the temporary label given to detached punctuation nodes is substituted by the label given to punctuation nodes.
	 * <p>
	 * Requires <code>flattenUnaryChains</code>.
	 * <p>
	 * TODO: Get the label for punctuation nodes from Properties.
	 * TODO: It assumes <code>flattenUnaryChains</code> has been run. I'd like to not need this.
	 * 
	 * @param tree  the parse tree
	 */
	private void moveCoordinationComma(Tree tree) {
//		System.out.println("\nMove comma in coordination with CONJ...");
		try {
			TregexPattern tregex = TregexPattern.compile("DETACHPNT=p >>- (__=l >, (__=t <- (__=r <, __=m <- (__ <, CONJ|PNT <- __=z))))");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
//				System.out.println("\nIntermediate:\n" + tree);
				Tree topNode = matcher.getNode("t");
				Tree leftNode = matcher.getNode("l");
				Tree rightNode = matcher.getNode("r");
				Tree punctNode = matcher.getNode("p");
				Tree middleNode = matcher.getNode("m");
				Tree lastNode = matcher.getNode("z");

				// Anchors
				Tree topAnchor = topNode;
				Tree leftAnchor = leftNode;
				Tree rightAnchor = rightNode;
				
				// TODO: Labels must be explicitly compared as a work-around for a bug in Tregex. After the bug is fixed, these checks should be enforced in the pattern
				String leftLabel = leftNode.label().toString().replaceFirst("[-*].+", "").replaceFirst("[P']$", "").replaceFirst("^SNS$", "S");
				String middleLabel = middleNode.label().toString().replaceFirst("[-*].+", "").replaceFirst("[P']$", "").replaceFirst("^SNS$", "S");
				String lastLabel = lastNode.label().toString().replaceFirst("[-*].+", "").replaceFirst("[P']$", "").replaceFirst("^SNS$", "S");

				if (leftLabel.equals(middleLabel) && middleLabel.equals(lastLabel) && leftNode != rightNode) {
//					System.out.println("\nT: " + topNode + "\nL: " + leftNode + "\nR: " + rightNode + "\nP: " + punctNode + "\nM: " + middleNode + "\nZ: " + lastNode);
					
					// Remove punctuation node
					Tree punctuationParent = punctNode.parent(tree);
					punctuationParent.removeChild(punctuationParent.indexOf(punctNode));
					
					// Create new node
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					List<Tree> newNodeLeaves = new ArrayList<Tree>(2);
					punctNode.label().setFromString("PNT");
					newNodeLeaves.add(punctNode);
					newNodeLeaves.add(rightAnchor);
					Tree newNode = treeFactory.newTreeNode(rightAnchor.label().toString().replaceFirst("\\*[^*]*\\*$", ""), newNodeLeaves);
					
					// Reassign children of top node
					Tree[] newTopChildren = {leftAnchor, newNode};
					topAnchor.setChildren(newTopChildren);
					
					// Since the tree has been changed (i.e. there could be new matches), reset the matcher
					matcher.reset();
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Normalizes coordination structures where the last element is marked by a punctuation symbol instead of a conjunction.
	 * <p>
	 * The method that moves punctuation symbols into their place in coordination structures (<code>moveCoordinationComma</code>) expects the final element in the coordination to be explicitly marked by a conjunction.
	 * However, in some cases (usually news headlines), the final element in the coordination is marked just by a punctuation symbol.
	 * This method moves the final punctuation symbol to the position that would be occupied by the conjunction so that <code>moveCoordinationComma</code> can then run unaltered.
	 * <p>
	 * This method requires that <code>detachPunctuation</code> be previously run to detach punctuation symbols from words.<br>
	 * Note also that, in some cases, this method relies on an unary chain to find the relevant structure. Accordingly, it has to run before <code>flattenUnaryChains</code>.
	 * <p>
	 * This method looks for the following structure:<br>
	 * (T (L ... (DETACHPNT ,)) (R (Z ...)))<br>
	 * where T=R, L=Z and R is not a modifier.
	 * <p>
	 * The punctuation node is moved into the position where the conjunction would be found if it existed.
	 * That is, it is placed as the first child of node R (i.e. as a left sister of node Z).
	 * Its label (the temporary label given to detached punctuation nodes) is replaced by the label given to punctuation nodes.
	 * <p>
	 * TODO: Get the label for punctuation nodes from Properties
	 *  
	 * @param tree  the parse tree
	 */
	private void moveNoConjCoordinationComma(Tree tree) {
//		System.out.println("\nMove comma in coordination w/out an explicit CONJ...");
		try {
			TregexPattern tregex = TregexPattern.compile("DETACHPNT=p >>- (__=l >1 (__=t <2 (__=r <: __=z)))");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
//				System.out.println("\nIntermediate:\n" + tree);
				Tree topNode = matcher.getNode("t");
				Tree leftNode = matcher.getNode("l");
				Tree rightNode = matcher.getNode("r");
				Tree punctNode = matcher.getNode("p");
				Tree lastNode = matcher.getNode("z");
				
				// Anchors
				Tree topAnchor = topNode;
				Tree leftAnchor = leftNode;
				Tree rightAnchor = rightNode;
				
				// Make some adjustments to the node selection
//				System.out.println("\nPerform selection adjustments, if needed...");
				if (leftNode.isUnaryRewrite() && !leftNode.label().toString().startsWith("SNS")) {
//					System.out.println("Adjust left (l): " + leftNode.label() + " -> " + leftNode.getChild(0).label());
					leftNode = leftNode.getChild(0);
				}
				if (lastNode.isUnaryRewrite() && !lastNode.label().toString().startsWith("SNS")) {
//					System.out.println("Adjust last (z): " + lastNode.label() + " -> " + lastNode.getChild(0).label());
					lastNode = lastNode.getChild(0);
				}
				if (topNode.parent(tree) != null && topNode.parent(tree).isUnaryRewrite()) {
//					System.out.println("Adjust top (t): " + topNode.label() + " -> " + topNode.parent(tree).label());
					topNode = topNode.parent(tree);
				}
				
				// Get node labels to check for equality
				// .replaceFirst("[-*].+", "").replaceFirst("[P']$", "").replaceFirst("^SNS$", "S");
				String leftLabel = leftNode.label().toString().replaceFirst("\\*(?:CP|GAP|MWU|ARG[0-9A])\\*", "").replaceFirst("-.+", "").replaceFirst("SNS", "S");
				String lastLabel = lastNode.label().toString().replaceFirst("\\*(?:CP|GAP|MWU|ARG[0-9A])\\*", "").replaceFirst("-.+", "").replaceFirst("SNS", "S");
				String topLabel = topNode.label().toString().replaceFirst("\\*(?:CP|GAP|MWU|ARG[0-9A])\\*", "").replaceFirst("-.+", "").replaceFirst("SNS", "S");
				String rightLabel = rightNode.label().toString().replaceFirst("\\*(?:CP|GAP|MWU|ARG[0-9A])\\*", "").replaceFirst("-.+", "").replaceFirst("SNS", "S");
				
				boolean isModifier = rightNode.label().toString().contains("*M*");

				if (leftLabel.equals(lastLabel) && topLabel.equals(rightLabel) && !isModifier) {
//					System.out.println("\nSelected tree:\n" + tree);
//					System.out.println("\nT: " + topNode + "\nL: " + leftNode + "\nR: " + rightNode + "\nP: " + punctNode + "\nZ: " + lastNode);
//					System.out.println("\nAnchor (t): " + topAnchor + "\nAnchor (l): " + leftAnchor + "\nAnchor (r): " + rightAnchor);
					
					// Remove punctuation node
					Tree punctuationParent = punctNode.parent(tree);
					punctuationParent.removeChild(punctuationParent.indexOf(punctNode));
					
					// Create new node
//					LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
					List<Tree> newNodeLeaves = new ArrayList<Tree>(2);
					punctNode.label().setFromString("PNT");
					newNodeLeaves.add(punctNode);
					newNodeLeaves.add(rightAnchor);
					Tree newNode = treeFactory.newTreeNode(rightAnchor.label().toString().replaceFirst("\\*[^*]*\\*$", ""), newNodeLeaves);
					
					// Reassign children of top node
					Tree[] newTopChildren = {leftAnchor, newNode};
					topAnchor.setChildren(newTopChildren);
					
					// Since the tree has been changed (i.e. there could be new matches), reset the matcher
					matcher.reset();
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	
	/*
	 * Methods that add nodes to the tree
	 */
	
	
	/**
	 * Inserts a XP node between a YP-* node and a single pre-terminal X (X!=Y) node it immediately dominates.
	 * <p>
	 * (YP-* (X leaf)) -> (YP-* (XP (X leaf)))
	 * <p>
	 * A phrasal node is not added if the top node has markings for empty categories.
	 * 
	 * @param tree
	 */
	private void addPhrasalNodes(Tree tree) {
//		System.out.println("\nAdd phrasal nodes...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^([^-\\/]+)P[^*]*$/#1%y=t  <: (/^([^-\\/])+/#1%x=c <: (__ !< __))  !<: /^([^-\\/])+/#1%y");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				Tree childNode = matcher.getNode("c");
				String newLabel = childNode.label().toString().replaceFirst("^([^-/]+)", "$1P");
//				LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
				topNode.setChildren(new Tree[]{treeFactory.newTreeNode(newLabel, topNode.getChildrenAsList())});
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Adds branches representing NULL subjects.
	 * <p>
	 * It requires that <code>annotateNullSubjectNodes</code> be previously run to annotate null subject nodes.
	 * <p>
	 * TODO: Assign ARGx according to the annotation added to the SNS node.
	 * TODO: Get the leaf name for null nodes from Properties.
	 *  
	 * @param tree
	 */
	private void addNullSJBranch(Tree tree) {
//		System.out.println("\nAdd NULL subject branch...");
		try {
			TregexPattern tregex = TregexPattern.compile("/^SNS/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree snsNode = matcher.getMatch();
				String snsLabel = snsNode.label().toString();
				if (snsLabel.matches(".+\\*(?:ARG.)?\\*$")) {
					String snsMark = snsLabel.replaceFirst("^.+\\*(ARG.)?\\*$", "$1");
					String newLabel = snsLabel.replaceFirst("^(.+)\\*(ARG.)?\\*$", "$1");
					snsNode.label().setFromString(newLabel);
					if (snsMark.length() > 0) {
						snsMark = "-" + snsMark;
					}
					Tree newBranch = string2tree("(NP-SJ" + snsMark + " *NULL*)");
					snsNode.addChild(0, newBranch);
				}
				snsNode.label().setFromString(snsNode.label().toString().replaceFirst("^SNS", "S"));
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Adds branches representing verb ellipsis.
	 * <p>
	 * Nodes are of the form (VP (X-PRD ...)) where X is not a V|VP.
	 * <p>
	 * TODO: Get the leaf name for ellipsis nodes from Properties.
	 * 
	 * @param tree
	 */
	private void addVerbEllipsisBranch(Tree tree) {
//		System.out.println("\nAdd verb ellipsis branch...");
		Tree newBranch = string2tree("(V *ELLIPSIS*)");
		try {
			TregexPattern tregex = TregexPattern.compile("@VP=t <: /^.+-PRD/ !<: /^VP?/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				topNode.addChild(0, newBranch);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Adds branches representing the ellipsis of constituents.
	 * <p>
	 * This method requires that <code>annotateEllipsisNodes</code> be previously run to identify ellipsis nodes and to mark whether the ellipsis branch should be a left child or a right child of the node.
	 * <p>
	 * The label of the new branch is based on the label of its parent node:
	 * <ul>
	 * <li>XP-* -> X
	 * <li>XP   -> X
	 * <li>X'   -> X
	 * </ul> 
	 * <p>
	 * TODO: Get the label for ellipsis nodes from Properties
	 * 
	 * @param tree
	 */
	private void addEllipsisBranch(Tree tree) {
//		System.out.println("\nAdd ELLIPSIS branch...\n" + tree);
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*(LEFT|RIGHT)\\*/=t");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				boolean addToLeft = topNode.label().toString().contains("*LEFT*");
				topNode.label().setFromString(topNode.label().toString().replaceFirst("\\*(LEFT|RIGHT)\\*", ""));
				String newLabel = topNode.label().toString().replaceFirst("^(.+?)(?:P-.+|P|')(?:\\*[^*]+\\*)*$", "$1");
				Tree newBranch = string2tree("(" + newLabel + " *ELLIPSIS*)");
				if (addToLeft) {
					topNode.addChild(0, newBranch);
				} else {
					topNode.addChild(newBranch);
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	
	/**
	 * A pre-processing fix for *GAP* nodes that is run by <code>addGapBranch</code>. 
	 *  
	 * Nodes marked with *GAP* are unary. However, in some rare cases, the *GAP* node and its child have the same category.
	 * In such cases, <code>flattenUnaryChains</code> excises the child node which might lead to a situation where the *GAP* node has more than one child and, as a consequence, the branch that will be subsequently added by <code>addGapBranch</code> will create a third child.
	 * This method re-inserts the node that was excised (for the excising to happen, its category must have been the same as the *GAP* node).
	 *
	 * TODO: To avoid the need for this fix, we will need to apply <code>flattenUnaryChains</code> differently (but how does this impact other transformations?). Either at a later time, as one of the final steps; or allow it to excise only unary chains leading to terminal nodes. 
	 * 
	 * @param tree
	 */
	private void preprocessGapBranch(Tree tree) {
//		System.out.println("\nPreprocess GAP branch...");
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*GAP[lr]\\*/=g !<: __");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree gapNode = matcher.getNode("g");
//				LabeledScoredTreeFactory treeFactory = new LabeledScoredTreeFactory(new StringLabelFactory());
				Tree[] newChildren = {treeFactory.newTreeNode(gapNode.label().toString().replaceFirst("-.+", ""), gapNode.getChildrenAsList())};
				gapNode.setChildren(newChildren);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Adds branches representing gapped (slashed) constituents.
	 * 
	 * @param tree
	 */
	private void addGapBranch(Tree tree) {
//		System.out.println("\nAdd GAP branch...");
		preprocessGapBranch(tree);
//		System.out.println(tree);
		int coindex = 1;
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*GAP[lr]\\*/=g");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				
				// Find node where the *GAP* branch is to be added
				Tree gapNode = matcher.getNode("g");
//				System.out.println("Found: " + gapNode);
				boolean addToLeft = gapNode.label().toString().contains("*GAPl*");
				gapNode.label().setFromString(gapNode.label().toString().replaceAll("\\*GAP[lr]\\*", ""));
				String slashLabel = gapNode.label().toString().replaceFirst("^[^/]+/", "").replaceAll("\\*[^*]+\\*", "");
				Tree newBranch = string2tree("(" + slashLabel + " *GAP*)");
				if (addToLeft) {
					gapNode.addChild(0, newBranch);
				} else {
					gapNode.addChild(newBranch);
				}
//				System.out.println(tree);

				// Go up to the topmost connected slash
				Tree topSlash = newBranch;
				boolean finished = false;
				while (!finished) {
					Tree parent = topSlash.parent(tree);
					if (parent != null && parent.label().toString().contains("/" + slashLabel)) {
						topSlash = parent;
					} else {
						finished = true;
					}
				}
//				System.out.println("Topmost: " + topSlash);
				
				// Co-index the gap with the filler (the sister of the topmost slash; there should be only one) 
				List<Tree> topSlashSiblings = topSlash.siblings(tree);
				for (Tree sibling : topSlashSiblings) {
//					System.out.println("Check sibling of topmost slash: " + sibling);
					Pattern fillerPattern = Pattern.compile(slashLabel.replaceFirst("P-", "[P']?-").replaceFirst("-M$", "-.+") + "(?:_(\\d+))?");
					Matcher fillerMatcher = fillerPattern.matcher(sibling.label().toString());
//					System.out.println("Match against: " + fillerPattern.pattern());
					if (fillerMatcher.matches()) {
//						System.out.println("Filler: " + sibling);
						String coindexLabel = fillerMatcher.group(1);
						if (coindexLabel == null) {
							coindexLabel = Integer.toString(coindex);
							sibling.label().setFromString(sibling.label().toString() + "_" + coindexLabel);
							coindex++;
						}
						newBranch.label().setFromString(newBranch.label().toString() + "_" + coindexLabel);
						
						// Relabel slashes when label of filler does not exactly match slash (i.e. X/X' vs XP)
						boolean equalLabels = sibling.label().toString().contentEquals(newBranch.label().toString());
//						System.out.println("Need relabeling?: " + sibling.label().toString() + " vs. " + newBranch.label().toString() + " [" + !equalLabels + "]");
						if (!equalLabels) {
							String oldSlashLabel = topSlash.label().toString().replaceFirst("^[^/]+/", "");
							String newSlashLabel = sibling.label().toString().replaceFirst("_\\d+$", "");
							TregexPattern relabelTregex = TregexPattern.compile("/\\/" + oldSlashLabel + "/");
							TregexMatcher relabelMatcher = relabelTregex.matcher(topSlash);
							while (relabelMatcher.findNextMatchingNode()) {
								Tree relabelNode = relabelMatcher.getMatch();
//								System.out.println("Relabel: " + relabelNode + " : " + oldSlashLabel + " -> " + newSlashLabel);
								relabelNode.label().setFromString(relabelNode.label().toString().replaceFirst(oldSlashLabel, newSlashLabel));
							}
							newBranch.label().setFromString(newBranch.label().toString().replaceFirst(oldSlashLabel, newSlashLabel));
						}
					} else {
//						System.out.println("No filler found");
						newBranch.firstChild().label().setFromString("*TOUGH*");
					}
				} // end for
				
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}
	
	
	
	/*
	 * Methods that remove information from the tree
	 */

	/**
	 * Removes markings that were added.
	 * 
	 * @param tree
	 */
	private void removeAnnotation(Tree tree, String pattern) {
//		System.out.println("\nRemove annotations matching a pattern...");
		try {
			TregexPattern tregex = TregexPattern.compile("/\\*" + pattern + "\\*/");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree node = matcher.getMatch();
				if (!node.isLeaf()) {
					node.label().setFromString(node.label().toString().replaceAll("\\*" + pattern + "\\*", ""));
				}
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}


	/*
	 * A main method for quick testing
	 */


	/**
	 * Code to test a bug in Tregex involving variable groups.
	 * Cf. email to parser-support list (Sep 9, 2010) and reply by Chris Manning (Nov 8, 2010).
	 * 
	 * Trees for testing:
	 * (T (X (N (N Moe (PNT ,)))) (NP (X (N Curly)) (NP (CONJ and) (X (N Larry)))))
	 * (T (X (X (N Moe (PNT ,)))) (NP (X (N Curly)) (NP (CONJ and) (X (N Larry)))))
	 * (T (X (Y (N Moe (PNT ,)))) (NP (X (N Curly)) (NP (CONJ and) (X (N Larry)))))
	 * (T (X (Y (Y Moe (PNT ,)))) (NP (X (N Curly)) (NP (CONJ and) (X (N Larry)))))
	 */
	@SuppressWarnings("unused")
	private void testBug(Tree tree) {
		System.out.println("\nTest Tregex bug...");
		try {
//			TregexPattern tregex = TregexPattern.compile("PNT=p >>- (__=l >, (__=t <- (__=r <, __=m <- (__ <, CONJ <- __=z))))");
			TregexPattern tregex = TregexPattern.compile("PNT=p >>- (/(.+)/#1%var=l >, (__=t <- (__=r <, /(.+)/#1%var=m <- (__ <, CONJ <- /(.+)/#1%var=z))))");
			TregexMatcher matcher = tregex.matcher(tree);
			while (matcher.find()) {
				Tree topNode = matcher.getNode("t");
				Tree leftNode = matcher.getNode("l");
				Tree rightNode = matcher.getNode("r");
				Tree punctNode = matcher.getNode("p");
				Tree middleNode = matcher.getNode("m");
				Tree lastNode = matcher.getNode("z");
				System.out.println("\nT: " + topNode + "\nL: " + leftNode + "\nR: " + rightNode + "\nP: " + punctNode + "\nM: " + middleNode + "\nZ: " + lastNode);
			}
		} catch (ParseException e) {
			e.printStackTrace();
		}
	}

	
	/**
	 * A main method for quick debugging.
	 * 
	 * @param args
	 */
	public static void main(String args[]) {
		Properties properties = new Properties();
		properties.setProperty("_LEXICONFILENAME", "/home/jsilva/Projects/concatLexicon.tdl");
		TreeHandler th = new TreeHandler(properties);
		
		Tree tree1 = null;
		Tree tree2 = null;
		
		tree1 = th.string2tree("(NP-SJ (NP (NP (N (N (N gato)) (R m))) (PK qw) (QW as)))");
		tree2 = th.string2tree("(NP-DO (XP (VP (A (N (Y cão)) (T h))) (II o) (LR gb)))");
		System.out.println("Tree1: " + tree1);
		System.out.println("Tree2: " + tree2);
		System.out.println("Same?: " + th.sameStructure(tree1, tree2));
		
		tree1 = th.string2tree("(A (B (C (D (E e) (F f)))))");
		tree2 = th.string2tree("(A (B (C (D (F f)) (E e))))");
		System.out.println("Tree1: " + tree1);
		System.out.println("Tree2: " + tree2);
		System.out.println("Same?: " + th.sameStructure(tree1, tree2));
		
	}
}
